# Copyright (c) 2025, NVIDIA CORPORATION.  All rights reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import re
from collections import defaultdict
from dataclasses import dataclass
from typing import DefaultDict, Dict, List, Tuple


@dataclass
class EntitySpan:
    entity_type: str
    start_position: int
    end_position: int


JSON_SCHEMA = {
    "type": "array",
    "items": {
        "type": "object",
        "required": ["entity_type", "entity_text"],
        "properties": {
            "entity_type": {"type": "string"},
            "entity_text": {"type": "string"},
        },
    },
}

# Recommended PII entities to redact
PII_LABELS = [
    "medical_record_number",
    "location",
    "address",
    "ssn",
    "date_of_birth",
    "date_time",
    "name",
    "email",
    "customer_id",
    "employee_id",
    "phone_number",
    "ip_address",
    "credit_card_number",
    "user_name",
    "device_identifier",
    "bank_routing_number",
    "company_name",
    "unique_identifier",
    "biometric_identifier",
    "account_number",
    "certificate_license_number",
    "license_plate",
    "vehicle_identifier",
    "api_key",
    "password",
    "health_plan_beneficiary_number",
    "national_id",
    "tax_id",
    "url",
    "swift_bic",
    "cvv",
    "pin",
]


def get_system_prompt(pii_labels: List[str] = PII_LABELS) -> str:
    return (
        "You are an expert redactor. The user is going to provide you with "
        "some text. Please find all personally identifying information from "
        "this text. Return results according to this JSON schema: "
        f"{str(JSON_SCHEMA)}\nOnly return results for entities which actually "
        "appear in the text. It is very important that you return the  "
        "entity_text by copying it exactly from the input. Do not perform any "
        "modification or normalization of the text. The entity_type should be "
        f"one of these: {', '.join(pii_labels)}\n"
    )


def validate_entity(
    entity: Dict[str, str],
    text: str,
    min_length: int = 2,
) -> bool:
    """Validate entity"""

    if not validate_keys(entity):
        return False

    if entity["entity_text"] not in text:
        return False

    if len(entity["entity_text"]) < min_length:
        return False

    return True


def validate_keys(entity_dict: Dict) -> bool:
    """Validate that keys in entity dict match schema"""

    required_keys = JSON_SCHEMA["items"]["required"]

    entity_keys = list(entity_dict.keys())
    if len(entity_keys) != len(required_keys):
        return False

    for k in required_keys:
        if k not in entity_keys:
            return False

    return True


def redact(
    full_text: str,
    pii_entities: List[Dict[str, str]],
) -> str:
    """Redact given entities from the original text"""

    entity_spans = find_entity_spans(full_text, pii_entities)

    # Initialize an offset to track the changes in the text
    redacted_text = full_text
    offset = 0
    for entity in fix_overlaps(entity_spans):
        entity_type = entity.entity_type
        start_position = entity.start_position + offset
        end_position = entity.end_position + offset

        # Replace the entity value with its type
        replacement = "{{" + entity_type + "}}"
        redacted_text = (
            redacted_text[:start_position] + replacement + redacted_text[end_position:]
        )

        # Update the offset
        offset += len(replacement) - (end_position - start_position)

    return redacted_text


def find_entity_spans(text: str, entities: List[Dict[str, str]]) -> List[EntitySpan]:
    """
    Find the start and end indexes for each entity in the given text.

    Args:
        text (str): The input text string.
        entities (list): A list of entities, where each entity is a dictionary
            containing the entity text and its type.

    Returns:
        list: A list of EntitySpan objects, where each contains the entity
            type, start position, and end position.

    """
    result = []
    seen: DefaultDict[Tuple, int] = defaultdict(int)
    for entity in entities:
        entity_text = entity["entity_text"]
        entity_type = entity["entity_type"]
        if (entity_text, entity_type) in seen:
            continue

        matches = re.finditer(re.escape(entity_text), text, re.IGNORECASE)
        if matches:
            result.extend(
                [
                    EntitySpan(entity_type, match.start(), match.end())
                    for match in matches
                ]
            )
            seen[(entity_text, entity_type)] += 1

    return result


def fix_overlaps(spans: List[EntitySpan]) -> List[EntitySpan]:
    """Handle overlaps in entity spans"""

    if not spans:
        return spans

    # Sort spans and iterate to check for overlap
    sorted_spans = sorted(spans, key=lambda x: (x.start_position, -x.end_position))
    results = [sorted_spans[0]]

    for span in sorted_spans[1:]:
        if span.start_position < results[-1].end_position:
            # Skip candidate span if sub-interval (regardless of category)
            if span.end_position < results[-1].end_position:
                continue

            else:
                if span.entity_type == results[-1].entity_type:
                    # If overlapping spans have same type, then extend boundary
                    results[-1].end_position = span.end_position

                else:
                    # If overlapping spans have different type, add new span
                    results.append(
                        EntitySpan(
                            span.entity_type,
                            results[-1].end_position + 1,
                            span.end_position,
                        )
                    )
        else:
            # Add non-overlapping span
            results.append(span)

    return results



# Copyright (c) 2024, NVIDIA CORPORATION.  All rights reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import argparse
import json
import os
import random
from typing import Any, List

import dask
from docbuilder import LawQADownloader, LawQAIterator
from filters import FilterLowScores
from modifiers import CleanHTML
from openai import AsyncOpenAI
from synthetic_gen import SyntheticGenerator

from nemo_curator import AsyncOpenAIClient, ScoreFilter, SemDedup, Sequential
from nemo_curator.datasets import DocumentDataset
from nemo_curator.filters import WordCountFilter
from nemo_curator.modifiers.unicode_reformatter import UnicodeReformatter
from nemo_curator.modules.config import SemDedupConfig
from nemo_curator.modules.modify import Modify
from nemo_curator.utils.distributed_utils import get_client
from nemo_curator.utils.file_utils import expand_outdir_and_mkdir
from nemo_curator.utils.script_utils import ArgumentHelper

SCRIPT_DIR_PATH = os.path.dirname(os.path.abspath(__file__))
DATA_DIR = os.path.join(SCRIPT_DIR_PATH, "data")
TEMP_DIR = os.path.join(SCRIPT_DIR_PATH, "_temp")
CONFIG_DIR = os.path.join(SCRIPT_DIR_PATH, "config")
DATASET_URL = "https://huggingface.co/datasets/ymoslem/Law-StackExchange/resolve/main/law-stackexchange-questions-answers.json"


def pre_imports():
    import cudf  # noqa: F401


def random_split_rows(rows: List[Any], train_ratio: float, val_ratio: float, seed=42):
    """
    Randomly splits a list of rows into training, validation, and test sets.

    Args:
        rows: The list of rows to be split.
        train_ratio: The ratio of rows to be allocated for training.
        val_ratio: The ratio of rows to be allocated for validation.
        seed: The seed value for random shuffling.

    Returns:
        A tuple containing the training, validation, and test sets.
    """
    random.seed(seed)
    random.shuffle(rows)
    n = len(rows)
    train_size = int(train_ratio * n)
    val_size = int(val_ratio * n)
    train_rows = rows[:train_size]
    val_rows = rows[train_size : train_size + val_size]
    test_rows = rows[train_size + val_size :]
    return train_rows, val_rows, test_rows


def download_and_convert_to_jsonl() -> str:
    """
    Downloads the Law Q&A dataset dataset and converts it to JSONL format.

    Returns:
        str: The path to the JSONL file.
    """
    download_dir = os.path.join(DATA_DIR, "raw", "downloads")
    splits_dir = os.path.join(DATA_DIR, "raw", "splits")
    os.makedirs(download_dir, exist_ok=True)
    os.makedirs(splits_dir, exist_ok=True)
    # Download the dataset in raw format and convert it to JSONL.
    downloader = LawQADownloader(download_dir)
    raw_fp = downloader.download(DATASET_URL)

    iterator = LawQAIterator()
    rows = []

    for record in iterator.iterate(raw_fp):
        json_record = json.dumps(record, ensure_ascii=False)
        rows.append(json_record)

    # Randomly split the rows into train, validation, and test sets.
    train_rows, val_rows, test_rows = random_split_rows(rows, 0.8, 0.1)

    # Write the split rows to separate JSONL files.
    for split_name, split_rows in zip(
        ["train", "val", "test"], [train_rows, val_rows, test_rows]
    ):
        split_fp = os.path.join(splits_dir, f"law-qa-{split_name}.jsonl")

        with open(split_fp, "w") as f:
            for row in split_rows:
                f.write(row + "\n")

    return (
        os.path.join(splits_dir, "law-qa-train.jsonl"),
        os.path.join(splits_dir, "law-qa-val.jsonl"),
        os.path.join(splits_dir, "law-qa-test.jsonl"),
    )


def semantic_dedupe(dataset):
    """
    Perform semantic deduplication on the given dataset.

    Args:
        dataset: The input DocumentDataset.

    Returns:
        The deduplicated DocumentDataset.
    """
    # Clean up the temporary directory to ensure everything is clean.
    if os.path.isdir(TEMP_DIR):
        os.system(f"rm -rf {TEMP_DIR}")

    semdedup_config = SemDedupConfig.from_yaml(
        os.path.join(CONFIG_DIR, "sem_dedup_config.yaml")
    )
    expand_outdir_and_mkdir(semdedup_config.cache_dir)
    semdup = SemDedup(
        config=semdedup_config,
        input_column="text",
        id_column="id",
        perform_removal=True,
    )
    result = semdup(dataset)
    return result


def run_curation_pipeline(
    args: Any,
    input_dir: str,
) -> DocumentDataset:
    """
    Run the curation pipeline on the dataset.

    Args:
        args: Command-line arguments.
        input_dir: The path to the uncurated JSONL file.

    Returns:
        The resulting dataset.
    """
    orig_dataset = DocumentDataset.read_json(input_dir, backend="pandas")
    dataset = orig_dataset

    cpu_curation_steps = Sequential(
        [
            #
            # Modifications
            #
            # Clean the HTML tags from all the records.
            Modify(CleanHTML(), text_field="title"),
            Modify(CleanHTML(), text_field="question"),
            Modify(CleanHTML(), text_field="answer"),
            # Unify the text encoding to Unicode.
            Modify(UnicodeReformatter(), text_field="title"),
            Modify(UnicodeReformatter(), text_field="question"),
            Modify(UnicodeReformatter(), text_field="answer"),
            #
            # Filtering
            #
            # Filter out records based on the question or answer word counts.
            ScoreFilter(
                WordCountFilter(min_words=50, max_words=500),
                text_field="question",
                score_type=int,
            ),
            ScoreFilter(
                WordCountFilter(min_words=50, max_words=500),
                text_field="answer",
                score_type=int,
            ),
            ScoreFilter(
                FilterLowScores(score_threshold=0),
                text_field="question_score",
                score_type=bool,
            ),
            ScoreFilter(
                FilterLowScores(score_threshold=0),
                text_field="answer_score",
                score_type=bool,
            ),
        ]
    )

    # Run the CPU curation steps.
    dataset = cpu_curation_steps(dataset)

    # Define and run the GPU curation steps.
    if args.device == "gpu":
        # Create a text field comprised of the title, question, and answer.
        # This field is used for finding semantically similar records and deduping them.
        dataset.df["text"] = (
            dataset.df["title"]
            + "\n"
            + dataset.df["question"]
            + "\n"
            + dataset.df["answer"]
        )
        dataset.df = dataset.df.to_backend("cudf")
        gpu_curation_steps = Sequential(
            [
                semantic_dedupe,
            ]
        )

        dataset = gpu_curation_steps(dataset)
        # Delete the text field as it is no longer needed.
        del dataset.df["text"]
        dataset.df = dataset.df.to_backend("pandas")

    dataset = dataset.persist()
    df = dataset.to_pandas()
    orig_len = len(orig_dataset.df)
    new_len = len(df)

    return df, orig_len, new_len


def run_pipeline(args, jsonl_fp):
    """
    Run the curation pipeline.

    Args:
        args: The command-line arguments.
        jsonl_fp: The file path to the JSONL file.

    Returns:
        The file path to the final curated JSONL file.
    """
    # Disable synthetic data generation if the necessary arguments are not provided.
    if not args.synth_gen_endpoint:
        print(
            "No synthetic data generation endpoint provided. Skipping synthetic data generation."
        )
        args.synth_gen_rounds = 0
    if not args.synth_gen_model:
        print(
            "No synthetic data generation model provided. Skipping synthetic data generation."
        )
        args.synth_gen_rounds = 0
    if not args.api_key:
        print(
            "No synthetic data generation API key provided. Skipping synthetic data generation."
        )
        args.synth_gen_rounds = 0

    if args.synth_gen_rounds:
        print(
            f"Using {args.synth_gen_endpoint}/{args.synth_gen_model} for synthetic data generation."
        )

    synth_gen_ratio = args.synth_gen_ratio
    synth_gen_rounds = args.synth_gen_rounds
    synth_n_variants = args.synth_n_variants

    assert (
        synth_gen_ratio >= 0 and synth_gen_ratio <= 1
    ), "The synthetic generation ratio must be between 0 and 1."
    assert (
        synth_gen_rounds >= 0
    ), "The number of synthetic generation rounds must be a non-negative integer."
    assert (
        synth_n_variants >= 1
    ), "The number of synthetic variants must be a positive integer."

    if args.device == "gpu":
        backend = "cudf"
    else:
        backend = "pandas"

    out_dir_base = os.path.join(DATA_DIR, "curated")
    jsonl_filename = os.path.basename(jsonl_fp)

    # Create the synthetic data generator.
    llm_client = AsyncOpenAIClient(
        AsyncOpenAI(
            base_url=args.synth_gen_endpoint,
            api_key=args.api_key or "",
            timeout=args.api_timeout,
        )
    )
    synth_gen = SyntheticGenerator(
        llm_client,
        sdg_model=args.synth_gen_model,
        sdg_model_kwargs={
            "top_p": 0.7,
            "max_tokens": 1024,
            "seed": 1234,
        },
        reward_model="nvidia/nemotron-4-340b-reward",
        n_variants=synth_n_variants,
    )

    with dask.config.set({"dataframe.backend": backend}):
        dask_client = get_client(**ArgumentHelper.parse_client_args(args))

        if args.device == "gpu":
            dask_client.run(pre_imports)

        print(f"Running the initial curation pipeline on '{jsonl_fp}'...")
        dataset_df, n_rows_before, n_rows_after = run_curation_pipeline(args, jsonl_fp)
        print(
            f"After the initial curation, the dataset has {n_rows_after} records (originally {n_rows_before})."
        )

        for i in range(1, synth_gen_rounds + 1):
            print(
                "--------------------------------------------------------------------------------"
            )
            print(
                f"Running synthetic data generation -- round {i} (out of {synth_gen_rounds})..."
            )
            out_dir = out_dir_base + f"/round-{i}"
            os.makedirs(out_dir, exist_ok=True)
            # Save the base dataset to disk.
            dataset_df.to_json(
                f"{out_dir}/{jsonl_filename}", orient="records", lines=True
            )

            #
            # Synthetic data generation
            #
            synth_prefix = f"{os.path.splitext(jsonl_filename)[0]}-synth-round-{i}"
            out_dir = synth_gen.run(dataset_df, out_dir, synth_prefix, synth_gen_ratio)
            #
            # Curation of the combined real and synthetic data
            #
            dataset_df, n_rows_before, n_rows_after = run_curation_pipeline(
                args, out_dir
            )

            print(
                f"After round {i}, the dataset has {n_rows_after} records (originally {n_rows_before})."
            )

        dask_client.cancel(dask_client.futures, force=True)
        dask_client.close()

    final_out_path = f"{out_dir_base}/final/{jsonl_filename}"
    os.makedirs(os.path.dirname(final_out_path), exist_ok=True)
    dataset_df.to_json(final_out_path, orient="records", lines=True)
    return final_out_path


def main():
    parser = argparse.ArgumentParser()
    parser = ArgumentHelper(parser).add_distributed_args()
    parser.add_argument(
        "--synth-gen-endpoint",
        type=str,
        default="https://integrate.api.nvidia.com/v1",
        help="The API endpoint to use for synthetic data generation. Any endpoint compatible with the OpenAI API can be used.",
    )
    parser.add_argument(
        "--synth-gen-model",
        type=str,
        default="nvidia/nemotron-4-340b-instruct",
        help="The model from the provided API endpoint to use for synthetic data generation. Leave blank to skip synthetic data generation.",
    )
    parser.add_argument(
        "--synth-gen-ratio",
        type=float,
        default=0.001,  # Use 0.1% of the real data for synthetic data generation to keep LLM calls low.
        help="The ratio of synthetic data to real data to generate. Synthetic data generation will be skipped if the value is 0.",
    )
    parser.add_argument(
        "--synth-gen-rounds",
        type=int,
        default=1,
        help="How many rounds of synthetic data generation to run. Will be ignored if --synth-gen-ratio is 0.",
    )
    parser.add_argument(
        "--synth-n-variants",
        type=int,
        default=1,
        help="The number of synthetic variants to generate for each record.",
    )
    parser.add_argument(
        "--api-key",
        type=str,
        default=None,
        help="The API key to use for the synthetic data generation LLM client.",
    )
    parser.add_argument(
        "--api-timeout",
        type=int,
        default=120,
        help="The timeout value for API calls in seconds.",
    )

    args = parser.parse_args()
    # Limit the total number of workers to ensure we don't run out of memory.
    args.n_workers = min(args.n_workers, 8)
    # Don't use RMM to prevent models from running out of memory.
    args.set_torch_to_use_rmm = False

    # Prepare the download and JSONL directories.
    if not os.path.isdir(DATA_DIR):
        os.makedirs(DATA_DIR)

    train_fp, val_fp, test_fp = download_and_convert_to_jsonl()
    train_fp_curated = run_pipeline(args, train_fp)

    curated_dir = os.path.dirname(train_fp_curated)
    os.system(f"cp {val_fp} {curated_dir}")
    os.system(f"cp {test_fp} {curated_dir}")
    print(
        "--------------------------------------------------------------------------------"
    )
    print(f"Curated files are saved in '{curated_dir}'.")


if __name__ == "__main__":
    main()



# Copyright (c) 2025, NVIDIA CORPORATION.  All rights reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
import bz2
import os
import subprocess
import tarfile
from urllib.parse import urlparse

import pytest

from nemo_curator.download import (
    ResiliparseExtractor,
    TrafilaturaExtractor,
    download_and_extract,
)
from nemo_curator.download.arxiv import ArxivDownloader, ArxivExtractor, ArxivIterator
from nemo_curator.download.commoncrawl import (
    CommonCrawlWARCDownloader,
    CommonCrawlWARCExtractor,
    CommonCrawlWARCIterator,
    JusTextExtractor,
    ResiliparseExtractor,
    get_common_crawl_urls,
    get_stop_list_dict,
)
from nemo_curator.download.wikipedia import (
    WikipediaDownloader,
    WikipediaExtractor,
    WikipediaIterator,
)


class DummyLock:
    def __enter__(self):
        return self

    def __exit__(self, exc_type, exc_val, exc_tb):
        pass


class FakeCompletedProcess:
    def __init__(self):
        self.returncode = 0


def fake_run_success(cmd, stdout, stderr):
    return FakeCompletedProcess()


@pytest.fixture
def html_string():
    # Modified from https://github.com/chatnoir-eu/chatnoir-resiliparse/blob/abdf1966fb3cefe3e0790e510ab5cb1446f99a79/tests/resiliparse/extract/test_html2text.py
    html = """<!doctype html>
        <head>
            <title>My Title</title>
            <meta charset="utf-8">
            <style>* { margin: 0; }</style>
        </head>
        <body>
            <section id="wrapper">
                <nav>
                    <ul>
                        <li>Nav 1</li>
                        <li>
                            <p>Nav 2</p>
                            <ul>
                                <li><p>Nav 3</p></li>
                            </ul>
                        </li>
                    </ul>
                </nav>
                <main>
                    This is a sample paragraph. In it we write words.
                    These are stopwords: because did than has near we almost while what still.
                    <a href="#foo" hidden>bar</a>

                    <p>
                    This paragraph doesn't have many stopwords. Remove it.
                    <br>Let's keep this paragraph: either came does last new took taken making became from.
                    </p>

                    <button aria-hidden="true">Click here</button>
                    <input type="hidden" value="foo">
                    <input type="text" value="Some text" placeholder="Insert text">
                    <input type="text" placeholder="Insert text">
                    <img src="" alt="Some image">
                    <object data="" class="some-class hidden">Cannot display object</object>
                </main>
                <script language="vbscript" type="text/vbscript">MsgBox("Hello World!")</script>
                <noscript>Sorry, your browser doesn't support VB Script!</noscript>
                <div><div><div><footer id="global-footer">
                    Copyright (C) 2021 Foo Bar
                </footer></div></div></div>
            </section>
        </body>
    </html>"""
    return html


class TestDownload:
    def test_imports(self):
        from nemo_curator.download import (
            JusTextExtractor,
            ResiliparseExtractor,
            TrafilaturaExtractor,
            download_arxiv,
            download_common_crawl,
            download_wikipedia,
        )

        assert True

    @pytest.mark.skip(
        reason="This test is flaky due to calling out to an external service and should be fixed."
    )
    def test_incorrect_snapshot_order(self):
        with pytest.raises(ValueError):
            end_snapshot = "2021-04"
            start_snapshot = "2021-10"
            urls = get_common_crawl_urls(start_snapshot, end_snapshot)

    @pytest.mark.skip(
        reason="This test is flaky due to calling out to an external service and should be fixed."
    )
    def test_incorrect_snapshot_order_news(self):
        with pytest.raises(ValueError):
            end_snapshot = "2021-04"
            start_snapshot = "2021-10"
            urls = get_common_crawl_urls(start_snapshot, end_snapshot, news=True)

    def test_no_urls(self):
        with pytest.raises(ValueError):
            output_format = {
                "text": str,
            }
            download_and_extract(
                [],
                [],
                CommonCrawlWARCDownloader(download_dir="."),
                CommonCrawlWARCIterator(),
                CommonCrawlWARCExtractor(),
                output_format,
            )

    def test_url_path_mismatch(self):
        with pytest.raises(ValueError):
            output_format = {
                "text": str,
            }
            download_and_extract(
                ["one", "two", "three"],
                ["one"],
                CommonCrawlWARCDownloader(download_dir="."),
                CommonCrawlWARCIterator(),
                CommonCrawlWARCExtractor(),
                output_format,
            )


class TestWikipedia:
    def test_wikipedia_downloader_existing_file(self, tmp_path, monkeypatch):
        # Create a temporary directory and simulate an already-downloaded file.
        download_dir = tmp_path / "downloads"
        download_dir.mkdir()

        url = "https://en.wikipedia.org/dummy-file"
        parsed = urlparse(url)
        output_name = parsed.path[1:].replace("/", "-")  # "dummy-file"
        file_path = os.path.join(str(download_dir), output_name)

        # Write a dummy file to simulate an existing download.
        with open(file_path, "w") as f:
            f.write("existing content")

        downloader = WikipediaDownloader(str(download_dir), verbose=False)

        # Monkey-patch subprocess.run (should not be called since file exists).
        monkeypatch.setattr(subprocess, "run", fake_run_success)

        result = downloader.download(url)
        assert result == file_path

    def test_wikipedia_downloader_new_file(self, tmp_path, monkeypatch):
        download_dir = tmp_path / "downloads"
        download_dir.mkdir()

        url = "https://en.wikipedia.org/new-file"
        parsed = urlparse(url)
        output_name = parsed.path[1:].replace("/", "-")  # "new-file"
        file_path = os.path.join(str(download_dir), output_name)

        # Ensure the file does not exist.
        if os.path.exists(file_path):
            os.remove(file_path)

        downloader = WikipediaDownloader(str(download_dir), verbose=False)
        downloader._lock = DummyLock()

        called_run = False

        def fake_run(cmd, stdout, stderr):
            nonlocal called_run
            called_run = True

            return FakeCompletedProcess()

        monkeypatch.setattr(subprocess, "run", fake_run)

        result = downloader.download(url)
        assert result == file_path
        assert called_run

    def test_wikipedia_iterator(self, tmp_path):
        # Create a minimal valid XML resembling a Wikipedia dump with one page.
        xml_content = """<?xml version="1.0"?>
<mediawiki xmlns="http://www.mediawiki.org/xml/export-0.10/">
  <page>
      <title>Test Article</title>
      <ns>0</ns>
      <id>123</id>
      <revision>
          <text>Test content with [[link]]</text>
      </revision>
  </page>
</mediawiki>"""
        # Compress the XML content using bz2.
        compressed_data = bz2.compress(xml_content.encode("utf-8"))

        # Write the compressed data to a temporary file.
        temp_file = tmp_path / "test_wiki.xml.bz2"
        temp_file.write_bytes(compressed_data)

        iterator = WikipediaIterator(language="en")
        pages = list(iterator.iterate(str(temp_file)))

        assert len(pages) == 1
        metadata, raw_text = pages[0]
        assert metadata["title"] == "Test Article"
        assert metadata["id"] == "123"
        # The URL is constructed by quoting the title.
        expected_url = "https://en.wikipedia.org/wiki/Test%20Article"
        assert metadata["url"] == expected_url
        assert "Test content with" in raw_text

    def test_wikipedia_extractor(self):
        extractor = WikipediaExtractor(language="en")
        # Sample wiki markup; note the presence of a heading and a magic word.
        content = "== Heading ==\nThis is a sample article. __NOTOC__"
        result = extractor.extract(content)

        # # The extractor should return a dict with a "text" key.
        assert isinstance(result, dict)
        extracted_text = result.get("text", "")
        # Verify that the magic word was removed.
        assert "__NOTOC__" not in extracted_text
        # Verify that the main content appears.
        assert "This is a sample article." in extracted_text


class TestArxiv:
    def test_arxiv_downloader_existing_file(self, tmp_path, monkeypatch):
        # Create a temporary download directory and simulate an already-downloaded tar file.
        download_dir = tmp_path / "downloads"
        download_dir.mkdir()
        tar_filename = "dummy.tar"
        file_path = os.path.join(str(download_dir), tar_filename)
        # Write dummy content to simulate an existing download.
        with open(file_path, "w") as f:
            f.write("existing content")

        downloader = ArxivDownloader(str(download_dir), verbose=False)
        # Monkey-patch subprocess.run (should not be called since file exists).
        monkeypatch.setattr(subprocess, "run", fake_run_success)
        result = downloader.download(tar_filename)
        assert result == file_path

    def test_arxiv_downloader_new_file(self, tmp_path, monkeypatch):
        # Create a temporary download directory and ensure the tar file does not exist.
        download_dir = tmp_path / "downloads"
        download_dir.mkdir()
        tar_filename = "dummy.tar"
        file_path = os.path.join(str(download_dir), tar_filename)
        if os.path.exists(file_path):
            os.remove(file_path)

        downloader = ArxivDownloader(str(download_dir), verbose=False)
        called_run = False

        def fake_run(cmd, stdout, stderr):
            nonlocal called_run
            called_run = True
            return FakeCompletedProcess()

        monkeypatch.setattr(subprocess, "run", fake_run)
        result = downloader.download(tar_filename)
        assert result == file_path
        assert called_run

    def test_arxiv_iterator(self, tmp_path):
        # Create an inner tar archive containing a .tex file.
        inner_tar_path = tmp_path / "2103.00001.tar"
        dummy_tex_filename = "2103.00001.tex"
        dummy_tex_content = "This is a dummy LaTeX content."
        with tarfile.open(inner_tar_path, "w") as inner_tar:
            # Create a temporary tex file to add into the inner tar archive.
            temp_tex_path = tmp_path / dummy_tex_filename
            with open(temp_tex_path, "w") as f:
                f.write(dummy_tex_content)
            inner_tar.add(temp_tex_path, arcname=dummy_tex_filename)

        # Create an outer tar archive that contains the inner tar archive.
        outer_tar_path = tmp_path / "dummy_main.tar"
        with tarfile.open(outer_tar_path, "w") as outer_tar:
            outer_tar.add(inner_tar_path, arcname="2103.00001.tar")

        iterator = ArxivIterator(log_frequency=1)
        results = list(iterator.iterate(str(outer_tar_path)))
        # Expect one paper extracted.
        assert len(results) == 1
        metadata, tex_files = results[0]
        # The ArxivIterator extracts the arxiv id from the inner archive's filename.
        assert metadata["id"] == "2103.00001"
        # The source_id is set to the outer tar file's basename.
        assert metadata["source_id"] == "dummy_main.tar"
        # Verify that the tex extraction returns the dummy content.
        assert isinstance(tex_files, list)
        assert dummy_tex_content in tex_files[0]

    def test_arxiv_extractor(self):
        extractor = ArxivExtractor()
        # Create a minimal LaTeX document including comments and a section header.
        content = r"""
        % This is a comment line that should be removed.
        \section{Introduction}
        This is the introduction of the paper.
        % Another comment that should vanish.
        """
        result = extractor.extract([content])
        assert isinstance(result, dict)
        extracted_text = result.get("text", "")
        # Verify that comments have been removed.
        assert "% This is a comment" not in extracted_text
        # Verify that the section header content is retained.
        assert "Introduction" in extracted_text
        assert "This is the introduction" in extracted_text


class TestCommonCrawl:
    def test_common_crawl_downloader_existing_file(self, tmp_path, monkeypatch):
        # Create a temporary downloads directory and simulate an already-downloaded file.
        download_dir = tmp_path / "downloads"
        download_dir.mkdir()
        url = "http://dummy/commoncrawl.warc"
        parsed = urlparse(url)
        output_name = parsed.path[1:].replace("/", "-")  # "commoncrawl.warc"
        file_path = os.path.join(str(download_dir), output_name)
        # Write dummy content to simulate an existing download.
        with open(file_path, "w") as f:
            f.write("existing content")

        downloader = CommonCrawlWARCDownloader(
            str(download_dir), aws=False, verbose=False
        )

        # Monkey-patch subprocess.run to track if it gets called.
        called_run = False

        def fake_run(cmd, stdout, stderr):
            nonlocal called_run
            called_run = True
            return FakeCompletedProcess()

        monkeypatch.setattr(subprocess, "run", fake_run)

        result = downloader.download(url)
        assert result == file_path
        # Since the file already exists, no download should be attempted.
        assert not called_run

    def test_common_crawl_downloader_new_file(self, tmp_path, monkeypatch):
        # Create a temporary downloads directory; ensure the file does not exist.
        download_dir = tmp_path / "downloads"
        download_dir.mkdir()
        url = "http://dummy/commoncrawl.warc"
        parsed = urlparse(url)
        output_name = parsed.path[1:].replace("/", "-")  # "commoncrawl.warc"
        file_path = os.path.join(str(download_dir), output_name)
        if os.path.exists(file_path):
            os.remove(file_path)

        downloader = CommonCrawlWARCDownloader(
            str(download_dir), aws=False, verbose=False
        )

        called_run = False

        def fake_run(cmd, stdout, stderr):
            nonlocal called_run
            called_run = True
            return FakeCompletedProcess()

        monkeypatch.setattr(subprocess, "run", fake_run)

        result = downloader.download(url)
        assert result == file_path
        # Since the file did not exist, a download call (and subprocess.run) should have been made.
        assert called_run

    def test_common_crawl_iterator(self, tmp_path):
        # Create a minimal valid WARC file with a single "response" record.
        raw_warc_path = tmp_path / "dummy.warc"
        http_response = (
            "HTTP/1.1 200 OK\r\n"
            "Content-Type: text/html\r\n"
            "\r\n"
            "<html><body><p>Common Crawl test paragraph with some content.</p></body></html>\r\n"
        )
        http_response_bytes = http_response.encode("utf-8")
        content_length = len(http_response_bytes)
        warc_record = (
            (
                f"WARC/1.0\r\n"
                f"WARC-Type: response\r\n"
                f"WARC-Record-ID: <urn:uuid:1234>\r\n"
                f"WARC-Date: 2022-01-01T00:00:00Z\r\n"
                f"WARC-Target-URI: http://example.com\r\n"
                f"Content-Length: {content_length}\r\n"
                f"\r\n"
            ).encode("utf-8")
            + http_response_bytes
            + b"\r\n\r\n"
        )
        raw_warc_path.write_bytes(warc_record)

        iterator = CommonCrawlWARCIterator(log_frequency=1)
        records = list(iterator.iterate(str(raw_warc_path)))
        assert len(records) == 1
        meta, content = records[0]
        # Check that the URL from the header is captured.
        assert "example.com" in meta["url"]
        # Verify that the content includes our test paragraph.
        assert b"Common Crawl test paragraph" in content

    def test_common_crawl_extractor_justext(self):
        extractor = CommonCrawlWARCExtractor(algorithm=JusTextExtractor())
        html = (
            "<html><body><p>Common Crawl test paragraph for justext extractor. "
            "Four score and seven years ago our fathers brought forth on this continent a new nation, "
            "conceived in liberty, and dedicated to the proposition that all men are created equal.</p></body></html>"
        )
        content = html.encode("utf-8")
        result = extractor.extract(content)
        print(result)
        assert result is not None
        # The extracted text should include our test paragraph.
        assert "Common Crawl test paragraph for justext extractor." in result["text"]
        assert "language" in result

    def test_common_crawl_extractor_resiliparse(self):
        extractor = CommonCrawlWARCExtractor(algorithm=ResiliparseExtractor())
        html = (
            "<html><body><p>Common Crawl test paragraph for resiliparse extractor. "
            "Four score and seven years ago our fathers brought forth on this continent a new nation, "
            "conceived in liberty, and dedicated to the proposition that all men are created equal.</p></body></html>"
        )
        content = html.encode("utf-8")
        result = extractor.extract(content)
        print(result)
        assert result is not None
        assert (
            "Common Crawl test paragraph for resiliparse extractor." in result["text"]
        )
        assert "language" in result


class TestExtractor:
    def test_resiliparse_extract_text(self, html_string):
        algorithm = ResiliparseExtractor()
        stop_words = get_stop_list_dict()
        result = algorithm.extract_text(html_string, stop_words["ENGLISH"], "ENGLISH")

        expected = [
            "This is a sample paragraph. In it we write words. These are stopwords: because did than has near we almost while what still.",
            "Let's keep this paragraph: either came does last new took taken making became from.",
        ]

        assert result == expected

    def test_trafilatura_extract_text(self, html_string):
        algorithm = TrafilaturaExtractor(
            min_extracted_size=10,
            min_duplcheck_size=10,
            max_repetitions=1,
            deduplicate=True,
        )
        stop_words = get_stop_list_dict()
        result = algorithm.extract_text(html_string, stop_words["ENGLISH"], "ENGLISH")

        expected = [
            "Let's keep this paragraph: either came does last new took taken making became from.",
        ]

        assert result == expected

    @pytest.mark.parametrize(
        "extraction_algorithm", ["justext", "resiliparse", "trafilatura"]
    )
    def test_extract_thai_text(self, extraction_algorithm):
        thai_html = """<!doctype html>
            <head>
                <title>ชื่อเรื่องของฉัน</title>
            </head>
            <body>
                    <main>
                        นี่คือตัวอย่างย่อหน้า ในนั้นเราเขียนคำต่างๆ
                        เหล่านี้เป็นคำหยุด: เพราะว่า ทำ กว่า มี ใกล้ เรา เกือบจะ ขณะที่ อะไร ยังคง

                        <p>
                        ย่อหน้านี้ไม่มีคำหยุดมากนัก ลบออก
                        <br>เรามาเก็บย่อหน้าไว้ดังนี้: ไม่ว่าจะมาทำอะไรใหม่ ๆ ก็เกิดขึ้น เกิดขึ้นจาก
                        </p>

                    </main>
            </body>
        </html>"""

        if extraction_algorithm == "justext":
            algorithm = JusTextExtractor()
            expected = [
                "นี่คือตัวอย่างย่อหน้า ในนั้นเราเขียนคำต่างๆ\nเหล่านี้เป็นคำหยุด: เพราะว่า ทำ กว่า มี ใกล้ เรา เกือบจะ ขณะที่ อะไร ยังคง",
                "ย่อหน้านี้ไม่มีคำหยุดมากนัก ลบออก\nเรามาเก็บย่อหน้าไว้ดังนี้: ไม่ว่าจะมาทำอะไรใหม่ ๆ ก็เกิดขึ้น เกิดขึ้นจาก",
            ]
        elif extraction_algorithm == "resiliparse":
            algorithm = ResiliparseExtractor()
            expected = [
                "นี่คือตัวอย่างย่อหน้า ในนั้นเราเขียนคำต่างๆ เหล่านี้เป็นคำหยุด: เพราะว่า ทำ กว่า มี ใกล้ เรา เกือบจะ ขณะที่ อะไร ยังคง",
                "ย่อหน้านี้ไม่มีคำหยุดมากนัก ลบออก",
                "เรามาเก็บย่อหน้าไว้ดังนี้: ไม่ว่าจะมาทำอะไรใหม่ ๆ ก็เกิดขึ้น เกิดขึ้นจาก",
            ]
        elif extraction_algorithm == "trafilatura":
            algorithm = TrafilaturaExtractor()
            expected = [
                "ย่อหน้านี้ไม่มีคำหยุดมากนัก ลบออก",
                "เรามาเก็บย่อหน้าไว้ดังนี้: ไม่ว่าจะมาทำอะไรใหม่ ๆ ก็เกิดขึ้น เกิดขึ้นจาก",
                "ย่อหน้านี้ไม่มีคำหยุดมากนัก ลบออก",
                "เรามาเก็บย่อหน้าไว้ดังนี้: ไม่ว่าจะมาทำอะไรใหม่ ๆ ก็เกิดขึ้น เกิดขึ้นจาก",
            ]

        stop_words = get_stop_list_dict()
        result = algorithm.extract_text(thai_html, stop_words["THAI"], "THAI")

        assert result == expected

    @pytest.mark.parametrize(
        "extraction_algorithm", ["justext", "resiliparse", "trafilatura"]
    )
    def test_extract_chinese_text(self, extraction_algorithm):
        chinese_html = """<!doctype html>
            <head>
                <title>我的标题</title>
            </head>
            <body>
                    <main>
                        这是一个示例段落。我们在其中写下单词。

                        <p>
                        本段落没有太多停用词。请将其删除。
                        <br>让我们保留这一段：要么来了，要么最后来了，要么新来了，要么采取了行动。
                        </p>

                    </main>
            </body>
        </html>"""

        if extraction_algorithm == "justext":
            algorithm = JusTextExtractor()
            expected = [
                "这是一个示例段落。我们在其中写下单词。",
                "本段落没有太多停用词。请将其删除。\n让我们保留这一段：要么来了，要么最后来了，要么新来了，要么采取了行动。",
            ]
        elif extraction_algorithm == "resiliparse":
            algorithm = ResiliparseExtractor()
            expected = [
                "这是一个示例段落。我们在其中写下单词。",
                "本段落没有太多停用词。请将其删除。",
                "让我们保留这一段：要么来了，要么最后来了，要么新来了，要么采取了行动。",
            ]
        elif extraction_algorithm == "trafilatura":
            algorithm = TrafilaturaExtractor()
            expected = [
                "这是一个示例段落。我们在其中写下单词。",
                "本段落没有太多停用词。请将其删除。",
                "让我们保留这一段：要么来了，要么最后来了，要么新来了，要么采取了行动。",
            ]

        stop_words = get_stop_list_dict()
        result = algorithm.extract_text(chinese_html, stop_words["CHINESE"], "CHINESE")

        assert result == expected

    @pytest.mark.parametrize(
        "extraction_algorithm", ["justext", "resiliparse", "trafilatura"]
    )
    def test_extract_japanese_text(self, extraction_algorithm):
        japanese_html = """<!doctype html>
            <head>
                <title>私のタイトル</title>
            </head>
            <body>
                    <main>
                        これはサンプルの段落です。ここに単語を書き込みます。

                        <p>
                        この段落にはストップワードがあまりありません。削除してください。
                        <br>この段落を維持しましょう: どちらかが来て、最後に新しいものを取って、作成し、なったのです。
                        </p>

                    </main>
            </body>
        </html>"""

        if extraction_algorithm == "justext":
            algorithm = JusTextExtractor()
            expected = [
                "これはサンプルの段落です。ここに単語を書き込みます。",
                "この段落にはストップワードがあまりありません。削除してください。\nこの段落を維持しましょう: どちらかが来て、最後に新しいものを取って、作成し、なったのです。",
            ]
        elif extraction_algorithm == "resiliparse":
            algorithm = ResiliparseExtractor()
            expected = [
                "これはサンプルの段落です。ここに単語を書き込みます。",
                "この段落にはストップワードがあまりありません。削除してください。",
                "この段落を維持しましょう: どちらかが来て、最後に新しいものを取って、作成し、なったのです。",
            ]
        elif extraction_algorithm == "trafilatura":
            algorithm = TrafilaturaExtractor()
            expected = [
                "この段落にはストップワードがあまりありません。削除してください。",
                "この段落を維持しましょう: どちらかが来て、最後に新しいものを取って、作成し、なったのです。",
                "この段落にはストップワードがあまりありません。削除してください。",
                "この段落を維持しましょう: どちらかが来て、最後に新しいものを取って、作成し、なったのです。",
            ]

        stop_words = get_stop_list_dict()
        result = algorithm.extract_text(
            japanese_html, stop_words["JAPANESE"], "JAPANESE"
        )

        assert result == expected

    @pytest.mark.parametrize(
        "extraction_algorithm", ["justext", "resiliparse", "trafilatura"]
    )
    def test_extract_korean_text(self, extraction_algorithm):
        korean_html = """<!doctype html>
            <head>
                <title>내 제목</title>
            </head>
            <body>
                    <main>
                        이것은 샘플 문단입니다. 여기에 단어를 적습니다.
                        이것들은 불용어입니다: 왜냐하면, 했으므로, 보다, 가까이에, 우리, 거의, 동안, 무엇, 아직도.

                        <p>
                        이 문단에는 불용어가 많지 않습니다. 제거하세요.
                        <br>이 문단을 유지해 보겠습니다: 왔거나 마지막이거나 새로운 것이거나 가져갔거나 만들어지거나 되었거나에서 왔습니다.
                        </p>

                    </main>
            </body>
        </html>"""

        if extraction_algorithm == "justext":
            algorithm = JusTextExtractor()
            expected = [
                "이것은 샘플 문단입니다. 여기에 단어를 적습니다.\n이것들은 불용어입니다: 왜냐하면, 했으므로, 보다, 가까이에, 우리, 거의, 동안, 무엇, 아직도.",
                "이 문단에는 불용어가 많지 않습니다. 제거하세요.\n이 문단을 유지해 보겠습니다: 왔거나 마지막이거나 새로운 것이거나 가져갔거나 만들어지거나 되었거나에서 왔습니다.",
            ]
        elif extraction_algorithm == "resiliparse":
            algorithm = ResiliparseExtractor()
            expected = [
                "이것은 샘플 문단입니다. 여기에 단어를 적습니다. 이것들은 불용어입니다: 왜냐하면, 했으므로, 보다, 가까이에, 우리, 거의, 동안, 무엇, 아직도.",
                "이 문단에는 불용어가 많지 않습니다. 제거하세요.",
                "이 문단을 유지해 보겠습니다: 왔거나 마지막이거나 새로운 것이거나 가져갔거나 만들어지거나 되었거나에서 왔습니다.",
            ]
        elif extraction_algorithm == "trafilatura":
            algorithm = TrafilaturaExtractor()
            expected = [
                "이 문단에는 불용어가 많지 않습니다. 제거하세요.",
                "이 문단을 유지해 보겠습니다: 왔거나 마지막이거나 새로운 것이거나 가져갔거나 만들어지거나 되었거나에서 왔습니다.",
                "이 문단에는 불용어가 많지 않습니다. 제거하세요.",
                "이 문단을 유지해 보겠습니다: 왔거나 마지막이거나 새로운 것이거나 가져갔거나 만들어지거나 되었거나에서 왔습니다.",
            ]

        stop_words = get_stop_list_dict()
        result = algorithm.extract_text(korean_html, stop_words["KOREAN"], "KOREAN")

        assert result == expected



# Copyright (c) 2025, NVIDIA CORPORATION.  All rights reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from hashlib import md5

import pandas as pd
import pytest
from dask import dataframe as dd

from nemo_curator.datasets import DocumentDataset
from nemo_curator.modules import ExactDuplicates


@pytest.fixture(
    params=[False, pytest.param(True, marks=pytest.mark.gpu)], ids=["no gpu", "gpu"]
)
def exact_dedup_data(request):
    df = pd.DataFrame(
        {"id": [1, 2, 300, 4, -1], "text": ["abc", "aba", "abb", "aba", "abc"]}
    )
    df = dd.from_pandas(df, 2)
    if request.param:
        df = df.to_backend("cudf")
    return DocumentDataset(df)


@pytest.fixture
def exact_no_dedup_data(request):
    # A dataset with no exact duplicates
    df = pd.DataFrame({"id": [1, 2, 300], "text": ["abc", "aba", "abb"]})
    df = dd.from_pandas(df, 2)
    return DocumentDataset(df)


class TestExactDuplicates:
    def test_unsupported_hash(self):
        with pytest.raises(ValueError):
            ExactDuplicates(hash_method="sha256")

    @pytest.mark.parametrize("cache_result", [False, True])
    def test_dup(self, exact_dedup_data, cache_result, tmpdir):
        exact_dups = ExactDuplicates(
            id_field="id",
            text_field="text",
            hash_method="md5",
            cache_dir=tmpdir if cache_result else None,
        )
        duplicates = exact_dups.identify_duplicates(exact_dedup_data)
        deduplicated_ds = exact_dups.remove(exact_dedup_data, duplicates)
        deduplicated_ids_series = deduplicated_ds.df.to_backend("pandas").compute()[
            "id"
        ]
        output_deduplicated_ids = set(deduplicated_ids_series.tolist())
        assert (
            len(output_deduplicated_ids) == 3
            and 300 in output_deduplicated_ids
            and len({-1, 1}.intersection(output_deduplicated_ids)) == 1
            and len({2, 4}.intersection(output_deduplicated_ids)) == 1
        )

        duplicates_df = (
            duplicates.df.to_backend("pandas")
            .compute()
            .sort_values(by="id", ignore_index=True)
        )
        expected_df = pd.DataFrame(
            {
                "id": [1, -1] + [2, 4],
                "_hashes": [md5(b"abc").hexdigest()] * 2
                + [md5(b"aba").hexdigest()] * 2,
            }
        ).sort_values(by="id", ignore_index=True)
        pd.testing.assert_frame_equal(duplicates_df, expected_df, check_like=True)

    def test_no_dedup(self, exact_no_dedup_data):
        exact_dups = ExactDuplicates(
            id_field="id",
            text_field="text",
            hash_method="md5",
            perform_removal=True,
        )
        result_df = exact_dups(exact_no_dedup_data).df.compute().reset_index(drop=True)
        expected_df = exact_no_dedup_data.df.compute().reset_index(drop=True)

        pd.testing.assert_frame_equal(result_df, expected_df)



# Copyright (c) 2024, NVIDIA CORPORATION.  All rights reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from __future__ import annotations

import logging
import math
import os
import time
import warnings
from typing import List, Optional, Tuple, Union

import cudf
import dask_cudf
import numpy as np

from nemo_curator.datasets import DocumentDataset
from nemo_curator.log import create_logger
from nemo_curator.utils.distributed_utils import performance_report_if_with_ts_suffix
from nemo_curator.utils.fuzzy_dedup_utils.io_utils import check_empty_buckets


class LSH:
    """
    Performs LSH on a MinhashSignatures
    """

    def __init__(
        self,
        cache_dir: str,
        num_hashes: int,
        num_buckets: int,
        buckets_per_shuffle: int = 1,
        false_positive_check: bool = False,
        logger: Union[logging.LoggerAdapter, str] = "./",
        id_fields: Union[str, list] = "id",
        minhash_field: str = "_minhash_signature",
        profile_dir: Optional[str] = None,
    ):
        """
        Parameters
        ----------
        cache_dir: str
          Needs to be specified, will compute & write duplicate id, bucket pairs to cache directory.
        num_hashes: Length of minhash signature
        num_buckets: Number of bands/buckets to create from the minhash signature.
          Hashes_per_signature = num_hashes / num_buckets
        buckets_per_shuffle: Number of bands/buckets to shuffle concurrently.
          but might lead to memory pressures and related errors.
        false_positive_check: bool
          If True, writes out buckets in a format compatible with downstream false positive check.
        logger: Existing logger to log to, or a path to a log directory.
        id_field: Columns in the Dataset denoting document ID.
        minhash_field: Column in the Dataset denoting minhash signature.
        profile_dir: str, Default None
          If specified directory to write dask profile
        """
        self.num_hashes = num_hashes
        self.num_buckets = num_buckets
        self.id_fields = [id_fields] if isinstance(id_fields, str) else id_fields
        self.minhash_field = minhash_field
        self.buckets_per_shuffle = buckets_per_shuffle
        self.bucket_ranges = self._generate_bucket_ranges(
            self.num_buckets, self.num_hashes
        )
        self.buckets_as_int = false_positive_check

        if cache_dir is None:
            raise ValueError(
                "cache_dir for intermediate outputs is required for this stage"
            )
        self.cache_dir = cache_dir
        self.profile_dir = profile_dir

        if isinstance(logger, str):
            self._logger = create_logger(
                rank=0,
                log_file=os.path.join(logger, "LSH.log"),
                name="LSH",
            )
        else:
            self._logger = logger

    def _generate_bucket_ranges(
        self, num_buckets: int, num_hashes: int
    ) -> List[List[int]]:
        """
        Generates a list of indices for the minhash ranges given num_bands &
        num_hashes.
        eg: num_bands=3, num_hashes=6
        [[0, 1], [2, 3], [4, 5]]
        """
        minhashes_per_bucket = num_hashes // num_buckets

        bucket_ranges = [
            list(
                range(
                    bucket * minhashes_per_bucket, (bucket + 1) * minhashes_per_bucket
                )
            )
            for bucket in range(num_buckets)
        ]
        return bucket_ranges

    def minhash_to_buckets(
        self,
        df: cudf.DataFrame,
        bucket_ranges: List[List[int]],
    ) -> cudf.DataFrame:
        df2 = df[self.id_fields]
        for i, h in enumerate(bucket_ranges):
            indices = cudf.Series([h]).repeat(len(df2))
            df2[f"_bucket_{i}"] = f"b{i}_" + df[self.minhash_field].list.take(
                indices
            ).hash_values(method="md5")
        return df2

    def bucket_id_to_int(
        self,
        bucket_ddf: dask_cudf.DataFrame,
        bucket_col_name: str = "bucket_id",
        start_id: int = 0,
    ) -> Tuple[dask_cudf.DataFrame, int]:
        """
        Maps bucket ids to a contigious integer range from starting from start_id.
        """
        unique_bucket_df = (
            bucket_ddf[[bucket_col_name]]
            .map_partitions(lambda x: x.drop_duplicates(ignore_index=True))
            .persist()
        )
        end_bucket_id = len(unique_bucket_df) - 1 + start_id
        unique_bucket_df["bucket_int_id"] = np.uint64(1)
        unique_bucket_df["bucket_int_id"] = unique_bucket_df["bucket_int_id"].cumsum()
        unique_bucket_df["bucket_int_id"] = (
            unique_bucket_df["bucket_int_id"] - 1 + start_id
        )
        bucket_ddf = bucket_ddf.merge(unique_bucket_df, on=[bucket_col_name])
        bucket_ddf = bucket_ddf.drop(columns=[bucket_col_name])
        bucket_ddf = bucket_ddf.rename(columns={"bucket_int_id": "_bucket_id"})
        bucket_ddf["_bucket_id"] = bucket_ddf["_bucket_id"].astype(np.uint64)
        return (bucket_ddf, end_bucket_id)

    def _minhash_to_bucket_meta(
        self, df: dask_cudf.DataFrame
    ) -> Tuple[cudf.DataFrame, int]:
        meta = df._meta_nonempty[self.id_fields]
        meta[self.minhash_field] = [np.ones(self.num_hashes)] * len(meta)
        return self.minhash_to_buckets(meta, self.bucket_ranges)

    def lsh(
        self,
        write_path: str,
        df: dask_cudf.DataFrame,
    ) -> bool:
        """
        Computes hash buckets for the DataFrame and writes them as parquet files to the specified path.

        Parameters:
            - write_path (str): The directory path to write parquet files.
            - df (dask_cudf.DataFrame): The input DataFrame with minhashes to be bucketed.
        Returns:
            are_buckets_empty: True if buckets were empty (no duplicates found), False otherwise.
        """
        wrote_buckets = False
        are_buckets_empty = True

        meta = self._minhash_to_bucket_meta(df)
        df = df.map_partitions(
            self.minhash_to_buckets,
            bucket_ranges=self.bucket_ranges,
            meta=meta,
        )
        bucket_start_id = 0
        for i in range(0, self.num_buckets, self.buckets_per_shuffle):
            bucket_columns = [
                f"_bucket_{i}"
                for i in range(i, min(self.num_buckets, i + self.buckets_per_shuffle))
            ]
            df2 = df.melt(
                id_vars=self.id_fields,
                value_name="_bucket_id",
                value_vars=bucket_columns,
            )[self.id_fields + ["_bucket_id"]]

            df2 = df2.shuffle(
                on=["_bucket_id"],
                ignore_index=True,
                npartitions=max(1, 2 ** math.floor(math.log2(df2.npartitions))),
            ).map_partitions(lambda x: x[x["_bucket_id"].duplicated(keep=False)])

            df2 = df2.reset_index(drop=True)
            # Buckets to Int
            if self.buckets_as_int:
                df2, end_id = self.bucket_id_to_int(
                    df2, bucket_col_name="_bucket_id", start_id=bucket_start_id
                )
                # If bucketing return empty dataframe
                if end_id < bucket_start_id:
                    self._logger.info(
                        f"No duplicate documents found for buckets: {bucket_columns}"
                    )
                    continue
                bucket_start_id = end_id + 1
                are_buckets_empty = False

            wrote_buckets, are_buckets_empty = self._write_bucket_parquet(
                df2,
                write_path,
                wrote_buckets,
                are_buckets_empty,
                bucket_columns,
            )

        if are_buckets_empty:
            self._logger.info("No duplicate documents found during LSH")
            if os.path.exists(write_path):
                import shutil

                shutil.rmtree(write_path)

        return are_buckets_empty

    def _write_bucket_parquet(
        self,
        df: dask_cudf.DataFrame,
        write_path: str,
        wrote_buckets: bool,
        are_buckets_empty: bool,
        buckets_to_write: List[str],
    ) -> tuple[bool, bool]:
        """
        Utility function to write the bucketed data to parquet
        handling cases of overwriting and appending as needed.
        """
        if not wrote_buckets:
            if os.path.exists(write_path):
                warnings.warn(
                    f"Output path {write_path} already exists and will be overwritten"
                )
            df.to_parquet(write_path, write_index=False, overwrite=True)
        else:
            df.to_parquet(
                write_path,
                write_index=False,
                overwrite=are_buckets_empty,
                append=not are_buckets_empty,
                ignore_divisions=True,
            )
        # Only check if buckets written so far are empty
        if are_buckets_empty:
            are_buckets_empty = check_empty_buckets(write_path)
        wrote_buckets = True

        if are_buckets_empty:
            self._logger.info(
                f"No duplicate documents found for buckets: {buckets_to_write}"
            )
        else:
            self._logger.info(f"Wrote data for buckets: {buckets_to_write}")
        return wrote_buckets, are_buckets_empty

    def __call__(self, dataset: DocumentDataset) -> DocumentDataset:
        df = dataset.df

        write_path = os.path.join(self.cache_dir, "_buckets.parquet")
        t0 = time.time()
        with performance_report_if_with_ts_suffix(self.profile_dir, "lsh-profile"):
            empty_result = self.lsh(write_path=write_path, df=df)
        self._logger.info(
            f"Time taken for LSH = {time.time() - t0}s and output written at {write_path}"
        )

        if empty_result:
            return None

        buckets_df = dask_cudf.read_parquet(write_path, split_row_groups=False)
        return DocumentDataset(buckets_df)



# Copyright (c) 2024, NVIDIA CORPORATION.  All rights reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import importlib
from abc import ABC, abstractmethod
from typing import Any, List, Union


class SyntheticDataGenerator(ABC):
    """
    An abstract base class for synthetic data generator pipeline.

    This class serves as a template for creating specific synethtic
    data generation pipelines.
    """

    def __init__(self):
        super().__init__()
        self._name = self.__class__.__name__

    @abstractmethod
    def generate(self, llm_prompt: Union[str, List[str]]) -> Union[str, List[str]]:
        pass

    @abstractmethod
    def parse_response(self, llm_response: Union[str, List[str]]) -> Any:
        pass



# Copyright (c) 2024, NVIDIA CORPORATION.  All rights reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.


import logging
import os
from typing import Literal, Tuple

import cudf
import cupy as cp
import dask.dataframe as dd
import pandas as pd
import torch
from crossfit.backend.cudf.series import create_list_series_from_1d_or_2d_ar

L2_DIST_TO_CENT_COL = "l2_dist_to_cent"
COSINE_DIST_TO_CENT_COL = "cosine_dist_to_cent"


def normalize_embeddings_col_in_df(
    df: cudf.DataFrame, embedding_col: str
) -> cudf.DataFrame:
    tensor = torch.Tensor(get_array_from_df(df, embedding_col))
    normalized_tensor = tensor / torch.norm(tensor, dim=1, keepdim=True)
    df[embedding_col] = create_list_series_from_1d_or_2d_ar(
        cp.asarray(normalized_tensor), index=df.index
    )
    return df


def get_array_from_df(df: cudf.DataFrame, embedding_col: str) -> cp.ndarray:
    return df[embedding_col].list.leaves.values.reshape(len(df), -1)


def add_l2_cosine_dist_to_centroid(
    df: cudf.DataFrame, embedding_col: str, centroids: cp.ndarray
) -> cudf.DataFrame:
    """
    Computes the L2 distance to nearest centroid to each embedding in the DataFrame.
    Embeddings are normalized. For cosine we'll need to normalize the centroids as well.
    """
    normalized_embeddings = get_array_from_df(df, embedding_col)
    centroids_ar = centroids[df["nearest_cent"].values]
    dist_to_cents = cp.sqrt(cp.sum((normalized_embeddings - centroids_ar) ** 2, axis=1))
    df[L2_DIST_TO_CENT_COL] = dist_to_cents
    del centroids_ar

    centroids_norm = centroids / cp.linalg.norm(centroids, axis=1, keepdims=True)
    centroids_ar = centroids_norm[df["nearest_cent"].values]
    # We normalize the centroids as well
    cosine_similarities = cp.sum(normalized_embeddings * centroids_ar, axis=1)
    df[COSINE_DIST_TO_CENT_COL] = 1 - cosine_similarities
    return df


def pairwise_cosine_similarity(
    cluster_reps: torch.Tensor,
    device: Literal["cuda", "cpu"],
) -> Tuple[cp.ndarray, cp.ndarray]:
    """
    Compute pairwise cosine similarity between cluster items,
    then replace to diagonal with zeros to ignore self similarity
    """
    # Move to device
    cluster_reps = cluster_reps.to(device)
    # Compute pairwise cosine similarity
    pairwise_sim_matrix = torch.mm(cluster_reps, cluster_reps.T)
    del cluster_reps
    # Get upper triangular matrix
    assert pairwise_sim_matrix.shape[0] == pairwise_sim_matrix.shape[1]
    triu_sim_mat = torch.triu(pairwise_sim_matrix, diagonal=1)
    # Get max similarity and indices
    max_values_and_indices = torch.max(triu_sim_mat, dim=0)
    max_similarity = max_values_and_indices[0]
    max_indices = max_values_and_indices[1]
    return cp.asarray(max_similarity, dtype=cp.float32), cp.asarray(max_indices)


def pairwise_cosine_similarity_batched(
    cluster_reps: torch.Tensor,
    device: Literal["cuda", "cpu"],
    batch_size: int = 1024,
) -> Tuple[cp.ndarray, cp.ndarray]:
    """
    Computes pairwise cosine similarity between cluster items,
    then replace to diagonal with zeros to ignore self similarity.
    This function is useful for large clusters where the pairwise similarity matrix
    does not fit into memory.
    We use a batched approach to compute the pairwise similarity matrix in batches.
    Memory requirements are O(N*B) where N is the number of items in the cluster and B is the batch size
    instead of O(N^2) for the full matrix.
    """
    cluster_reps = cluster_reps.to(device)
    max_similarity = torch.zeros(
        cluster_reps.shape[0], dtype=torch.float32, device=device
    )
    max_indices = torch.zeros(cluster_reps.shape[0], dtype=torch.int64, device=device)
    for start_idx in range(0, cluster_reps.shape[0], batch_size):
        end_idx = min(start_idx + batch_size, cluster_reps.shape[0])
        batch = cluster_reps[start_idx:end_idx]
        pairwise_sim_matrix = torch.mm(cluster_reps, batch.T)
        triu_sim_matrix = torch.triu(pairwise_sim_matrix, diagonal=1 - start_idx)
        del batch, pairwise_sim_matrix
        max_values_and_indices = torch.max(triu_sim_matrix, dim=0)
        max_similarity[start_idx:end_idx] = max_values_and_indices[0]
        max_indices[start_idx:end_idx] = max_values_and_indices[1]

    return cp.asarray(max_similarity), cp.asarray(max_indices)


def get_semantic_matches_per_cluster(
    cluster_id: int,
    emb_by_clust_dir: str,
    id_col: str,
    output_dir: str,
    embedding_col: str,
    which_to_keep: Literal["hard", "easy", "random"],
    sim_metric: Literal["cosine", "l2"],
    batched_cosine_similarity: int = 1024,
) -> None:
    """
    Get the semantic matches for a single cluster.
    Reads the cluster embeddings and then computes pairwise cosine similarity between them.
    """
    if sim_metric == "cosine":
        distance_col = COSINE_DIST_TO_CENT_COL
    elif sim_metric == "l2":
        distance_col = L2_DIST_TO_CENT_COL
    else:
        msg = f"Invalid similarity metric: {sim_metric}. Only cosine and l2 are supported."
        raise ValueError(msg)

    cluster_df = cudf.read_parquet(
        os.path.join(emb_by_clust_dir, f"nearest_cent={cluster_id}"),
        columns=[embedding_col, id_col, distance_col],
    )
    output_df_file_path = os.path.join(output_dir, f"cluster_{cluster_id}.parquet")
    if len(cluster_df) == 1:
        cluster_df["id"] = cluster_df[id_col]
        cluster_df["max_id"] = cluster_df[id_col]
        cluster_df["cosine_sim_score"] = [0]
        cluster_df = cluster_df[["id", "max_id", "cosine_sim_score"]]
        cluster_df.to_parquet(output_df_file_path)
        return

    if which_to_keep == "hard":
        cluster_df = cluster_df.sort_values(
            by=[distance_col, id_col], ascending=False, ignore_index=True
        )
    elif which_to_keep == "easy":
        cluster_df = cluster_df.sort_values(
            by=[distance_col, id_col], ascending=True, ignore_index=True
        )
    elif which_to_keep == "random":
        cluster_df = cluster_df.sample(frac=1, random_state=42, ignore_index=True)

    cluster_embeddings = torch.as_tensor(
        get_array_from_df(cluster_df, embedding_col), device="cuda"
    )
    ids = cluster_df[id_col]
    assert cluster_embeddings.shape[0] == len(ids)

    if batched_cosine_similarity > 0:
        max_similarity, max_indices = pairwise_cosine_similarity_batched(
            cluster_embeddings, "cuda", batched_cosine_similarity
        )
    else:
        max_similarity, max_indices = pairwise_cosine_similarity(
            cluster_embeddings, "cuda"
        )
    max_indices_id = ids.iloc[max_indices].reset_index(drop=True)
    points_to_remove_df = cudf.DataFrame(
        {
            "id": ids,
            "max_id": max_indices_id,
            "cosine_sim_score": max_similarity,
        }
    )
    points_to_remove_df.to_parquet(output_df_file_path)


def prune_single_cluster(
    cluster_id: int,
    id_col: str,
    emb_by_clust_dir: str,
    semdedup_pruning_tables_dir: str,
    eps: float,
) -> cudf.DataFrame:
    """
    Processes data for a single cluster, applying pruning based on specified epsilon.

    Args:
        cluster_id (int): The specific cluster ID to process.
        id_col (str): The name of the ID column.
        emb_by_clust_dir (str): Path to where clustered embeddings are stored.
        semdedup_pruning_tables_dir (str): Path to the pruning tables directory.
        eps (float): Epsilon value for pruning.

    Returns:
        cudf.DataFrame: A DataFrame of the pruned cluster data
    """
    cluster_dir = os.path.join(emb_by_clust_dir, f"nearest_cent={cluster_id}")
    # For the output we only return id, cosine_dist_to_cent, and cluster
    df_cluster = cudf.read_parquet(
        cluster_dir, columns=[id_col, COSINE_DIST_TO_CENT_COL]
    ).assign(cluster=cluster_id)

    pruning_table_fname = os.path.join(
        semdedup_pruning_tables_dir, f"cluster_{cluster_id}.parquet"
    )
    # In future we can add more columns to the pruning table like max_id etc.
    pruning_table = cudf.read_parquet(
        pruning_table_fname, columns=["id", "cosine_sim_score"]
    )
    # If the pruning table only has one row, we don't need to remove any records
    if len(pruning_table) == 1:
        # Create empty dataframe with same schema / dtypes as df_cluster
        empty_df = cudf.DataFrame(columns=df_cluster.columns).astype(df_cluster.dtypes)
        return empty_df
    # We keep only records that are very similar i.e cosine_sim_score >= 1 - eps
    pruning_table = pruning_table[pruning_table["cosine_sim_score"] >= 1 - eps][["id"]]
    # In future we can avoid this merge if we add more columns to the pruning table
    # However that might increase memory consumption at that stage, keeping it as is for now
    return df_cluster.merge(
        pruning_table.rename(columns={"id": id_col}), on=id_col, how="inner"
    )


def write_pruned_summary_file(
    eps: float,
    emb_by_clust_dir: str,
    filtered_unique_ids_path: str,
    output_summary_file: str,
    logger: logging.Logger,
):
    """
    Writes a summary file for the pruned data.
    """
    removed = len(dd.read_parquet(filtered_unique_ids_path))
    total = len(dd.read_parquet(emb_by_clust_dir))
    kept = total - removed

    logger.info(
        f"DONE saving {kept} out of {total}. Removed: {removed}. Epsilon: {eps:.4f}"
    )
    result_dict = {
        "eps": [eps],
        "kept": [kept],
        "removed": [removed],
        "total": [total],
    }
    df = pd.DataFrame(result_dict)
    df.to_csv(output_summary_file, index=False)



# Copyright (c) 2024, NVIDIA CORPORATION.  All rights reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import sys

import dask
from packaging.version import parse as parse_version

try:
    _dask_version = parse_version(dask.__version__)
except TypeError:
    # When mocking with autodoc the dask version is not there
    _dask_version = parse_version("2024.06.0")


try:
    import dask_cudf

    _dask_cudf_version = parse_version(dask_cudf.__version__)
except (ImportError, TypeError):
    # When mocking with autodoc the dask version is not there
    _dask_cudf_version = parse_version("2024.06.0")

try:
    import cudf

    CURRENT_CUDF_VERSION = parse_version(cudf.__version__)
except (ImportError, TypeError):
    CURRENT_CUDF_VERSION = parse_version("24.10.0")

# TODO remove this once 25.02 becomes the base version of cudf in nemo-curator

# minhash in < 24.12 used to have a minhash(txt) api which was deprecated in favor of
# minhash(a, b) in 25.02 (in 24.12, minhash_permuted(a,b) was introduced)
MINHASH_DEPRECATED_API = (
    CURRENT_CUDF_VERSION.base_version < parse_version("24.12").base_version
)
MINHASH_PERMUTED_AVAILABLE = (CURRENT_CUDF_VERSION.major == 24) & (
    CURRENT_CUDF_VERSION.minor == 12
)

# TODO: remove when dask min version gets bumped
DASK_SHUFFLE_METHOD_ARG = _dask_version > parse_version("2024.1.0")
DASK_P2P_ERROR = _dask_version < parse_version("2023.10.0")
DASK_SHUFFLE_CAST_DTYPE = _dask_version > parse_version("2023.12.0")
DASK_CUDF_PARQUET_READ_INCONSISTENT_SCHEMA = _dask_version > parse_version("2024.12")

# Query-planning check (and cache)
_DASK_QUERY_PLANNING_ENABLED = None


def query_planning_enabled():
    global _DASK_QUERY_PLANNING_ENABLED

    if _DASK_QUERY_PLANNING_ENABLED is None:
        if _dask_version > parse_version("2024.12.1"):
            _DASK_QUERY_PLANNING_ENABLED = True
        elif _dask_version > parse_version("2024.6.0"):
            import dask.dataframe as dd

            _DASK_QUERY_PLANNING_ENABLED = dd.DASK_EXPR_ENABLED
        else:
            _DASK_QUERY_PLANNING_ENABLED = "dask_expr" in sys.modules
    return _DASK_QUERY_PLANNING_ENABLED



# Copyright (c) 2024, NVIDIA CORPORATION.  All rights reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import argparse
import pickle

import nemo_curator
from nemo_curator.datasets import DocumentDataset
from nemo_curator.utils.distributed_utils import get_client, read_data
from nemo_curator.utils.file_utils import get_all_files_paths_under
from nemo_curator.utils.script_utils import ArgumentHelper


def main(args):
    client = get_client(**ArgumentHelper.parse_client_args(args))

    # Each rank read in the task data
    with open(args.input_task_ngrams, "rb") as fp:
        task_ngrams = pickle.load(fp)

    decontaminator = nemo_curator.TaskDecontamination(
        [], text_field=args.input_text_field, max_ngram_size=args.max_ngram_size
    )

    files = get_all_files_paths_under(args.input_data_dir)
    dataset = DocumentDataset(
        read_data(files, file_type=args.input_file_type, backend="pandas")
    )

    result = decontaminator.find_matching_ngrams(task_ngrams, dataset).compute()
    print(f"Found a total of {len(result['matched-ngrams'])} matching n-grams")

    output = {
        "matched-ngrams": result["matched-ngrams"],
        "ngrams-freq": result["ngrams-freq"],
        "max-ngram-size": args.max_ngram_size,
        "min-ngram-size": args.min_ngram_size,
    }
    with open(args.output_matched_ngram_data, "wb") as fp:
        pickle.dump(output, fp)


def attach_args(
    parser=argparse.ArgumentParser(
        """
    Searches for matching task n-grams in the input dataset
    and writes out a list of n-grams that were found.
""",
        formatter_class=argparse.ArgumentDefaultsHelpFormatter,
    )
):
    argumentHelper = ArgumentHelper(parser)

    argumentHelper.add_arg_input_data_dir()
    argumentHelper.add_arg_input_file_type()
    argumentHelper.add_arg_input_text_field()
    argumentHelper.add_distributed_args()
    parser.add_argument(
        "--input-task-ngrams",
        type=str,
        default=None,
        help="",
    )
    parser.add_argument(
        "--max-ngram-size",
        type=int,
        default=13,
        help="The maximum n-gram size to consider within the dataset.",
    )
    parser.add_argument(
        "--min-ngram-size",
        type=int,
        default=8,
        help="The minimum n-gram size to consider within the datset.",
    )
    parser.add_argument(
        "--output-matched-ngram-data",
        type=str,
        default=None,
        help="Output dictionary that contains the output matched n-grams "
        "and the frequency of their matches, min-ngram size, max-ngram "
        "size, and the frequencies of n-gram sizes. All of these data will be "
        "used by remove_matching_grams for which this program is a prerequisite.",
    )

    return parser


def console_script():
    main(attach_args().parse_args())



# Copyright (c) 2025, NVIDIA CORPORATION.  All rights reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import argparse
import json
import os
from functools import partial
from typing import Any

from docbuilder import EmailsDownloader, EmailsIterator
from filters import FilterEmailsWithLongBody, FilterEmptyEmails
from modifiers import AddPeriod, AddSystemPrompt

from nemo_curator import ScoreFilter, Sequential
from nemo_curator.datasets import DocumentDataset
from nemo_curator.modifiers.pii_modifier import PiiModifier
from nemo_curator.modifiers.unicode_reformatter import UnicodeReformatter
from nemo_curator.modules.modify import Modify
from nemo_curator.utils.distributed_utils import get_client
from nemo_curator.utils.script_utils import ArgumentHelper

SCRIPT_DIR_PATH = os.path.dirname(os.path.abspath(__file__))
DATA_DIR = os.path.join(SCRIPT_DIR_PATH, "data")
DATASET_URL = "https://huggingface.co/datasets/neelblabla/enron_labeled_emails_with_subjects-llama2-7b_finetuning/raw/main/prompts_train.csv"


def download_and_convert_to_jsonl() -> str:
    """
    Downloads the emails dataset and converts it to JSONL format.

    Returns:
        str: The path to the JSONL file.
    """

    # Download the dataset in raw format and convert it to JSONL.
    downloader = EmailsDownloader(DATA_DIR)
    output_path = os.path.join(DATA_DIR, "emails.jsonl")
    raw_fp = downloader.download(DATASET_URL)

    iterator = EmailsIterator()

    # Parse the raw data and write it to a JSONL file.
    with open(output_path, "w") as f:
        for record in iterator.iterate(raw_fp):
            json_record = json.dumps(record, ensure_ascii=False)
            f.write(json_record + "\n")

    return output_path


def redact_pii(dataset: DocumentDataset, text_field) -> DocumentDataset:
    """
    Redacts personally identifiable information (PII) from a given dataset.

    Args:
        dataset (DocumentDataset): The dataset containing documents with PII.

    Returns:
        DocumentDataset: The redacted dataset with PII replaced by a generic value.
    """
    redactor = Modify(
        PiiModifier(
            supported_entities=[
                "ADDRESS",
                "EMAIL_ADDRESS",
                "LOCATION",
                "PERSON",
                "URL",
                "PHONE_NUMBER",
            ],
            anonymize_action="replace",
            device="cpu",
        ),
        text_field=text_field,
    )
    return redactor(dataset)


def run_curation_pipeline(args: Any, jsonl_fp: str) -> str:
    """
    Run the curation pipeline on the dataset.

    Args:
        args (Any): Command-line arguments.
        jsonl_fp (str): The path to the uncurated JSONL file.

    Returns:
        str: The path to the curated JSONL file.
    """
    client = get_client(**ArgumentHelper.parse_client_args(args))
    print(f"    Running the curation pipeline on '{jsonl_fp}'...")
    orig_dataset = DocumentDataset.read_json(jsonl_fp)
    dataset = orig_dataset

    redact_pii_subject = partial(redact_pii, text_field="subject")
    redact_pii_body = partial(redact_pii, text_field="body")

    curation_steps = Sequential(
        [
            #
            # Unify the text encoding to Unicode.
            #
            Modify(UnicodeReformatter(), text_field="subject"),
            Modify(UnicodeReformatter(), text_field="body"),
            Modify(UnicodeReformatter(), text_field="category"),
            #
            # Filtering
            #
            # Filter out empty emails.
            ScoreFilter(
                FilterEmptyEmails(), text_field="subject", score_type=bool, invert=True
            ),
            ScoreFilter(
                FilterEmptyEmails(), text_field="body", score_type=bool, invert=True
            ),
            ScoreFilter(
                FilterEmptyEmails(), text_field="category", score_type=bool, invert=True
            ),
            # Filter out emails that are too long.
            ScoreFilter(FilterEmailsWithLongBody(), text_field="body", score_type=bool),
            #
            # Redact personally identifiable information (PII).
            #
            redact_pii_subject,
            redact_pii_body,
            #
            # Final modifications.
            #
            # Add system prompts to every email, which helps the model focus on the task.
            Modify(AddSystemPrompt(), text_field="body"),
            # Add a period to the end of each email category, which makes PEFT easier.
            Modify(AddPeriod(), text_field="category"),
        ]
    )

    dataset = curation_steps(dataset)
    dataset = dataset.persist()

    print(f"    Original dataset length: {len(orig_dataset.df)}")
    print(f"    After running the curation pipeline: {len(dataset.df)}")
    print(f"    Writing to '{jsonl_fp}'...")
    out_path = os.path.join(
        os.path.dirname(jsonl_fp),
        "curated",
    )
    os.makedirs(out_path, exist_ok=True)
    dataset.to_json(out_path, write_to_filename=True)
    client.close()
    return os.path.join(out_path, os.path.basename(jsonl_fp))


def main():
    parser = argparse.ArgumentParser()
    args = ArgumentHelper(parser).add_distributed_args().parse_args()
    # Limit the total number of workers to ensure we don't run out of memory.
    args.n_workers = min(args.n_workers, 8)

    # Prepare the download and JSONL directories.
    if not os.path.isdir(DATA_DIR):
        os.makedirs(DATA_DIR)

    jsonl_fp = download_and_convert_to_jsonl()
    run_curation_pipeline(args, jsonl_fp)


if __name__ == "__main__":
    main()



# Copyright (c) 2025, NVIDIA CORPORATION.  All rights reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.


import os
import subprocess
import unicodedata
import warnings
from abc import ABC, abstractmethod
from copy import deepcopy
from typing import Literal, Optional
from urllib.parse import urlparse

import justext
import lxml
import pycld2 as cld2
from charset_normalizer import detect
from resiliparse.extract.html2text import extract_plain_text
from trafilatura import extract as extract_with_trafilatura
from trafilatura.settings import DEFAULT_CONFIG as TRAFILATURA_DEFAULT_CONFIG
from warcio.archiveiterator import ArchiveIterator

from nemo_curator.datasets import DocumentDataset
from nemo_curator.download.doc_builder import (
    DocumentDownloader,
    DocumentExtractor,
    DocumentIterator,
    download_and_extract,
)
from nemo_curator.utils.download_utils import get_common_crawl_urls
from nemo_curator.utils.file_utils import expand_outdir_and_mkdir

NON_SPACED_LANGUAGES = ["THAI", "CHINESE", "JAPANESE", "KOREAN"]


def decode_html(html_bytes):
    # Convert from bytes to text using utf-8 encoding
    try:
        return html_bytes.decode("utf-8")
    except UnicodeDecodeError:
        # If utf-8 fails, try to find a different encoding
        return try_decode_with_detected_encoding(html_bytes)


def try_decode_with_detected_encoding(html_bytes):
    detected_encoding = detect(html_bytes)["encoding"]
    bad_detection = not detected_encoding or detected_encoding == "utf-8"
    if bad_detection:
        return None
    try:
        return html_bytes.decode(detected_encoding)
    except:
        return None


def lang_detect(decoded_html):
    try:
        details = cld2.detect(decoded_html)[2]
    except Exception:
        # Remove control characters
        cleaned_html = "".join(
            i for i in decoded_html if unicodedata.category(i)[0] != "C"
        )
        details = cld2.detect(cleaned_html)[2]

    return details[0][0].upper()


class HTMLExtractorAlgorithm(ABC):
    @abstractmethod
    def extract_text(self, html, stop_words, language):
        pass


class JusTextExtractor(HTMLExtractorAlgorithm):
    def __init__(
        self,
        length_low=70,
        length_high=200,
        stopwords_low=0.30,
        stopwords_high=0.32,
        max_link_density=0.2,
        max_heading_distance=200,
        no_headings=False,
        is_boilerplate=None,
        logger=None,
    ):
        """
        Initialize the jusText text extraction algorithm with specified parameters.

        jusText is a tool for removing boilerplate content, such as navigation links, headers, and footers from HTML pages.
        It is designed to preserve mainly text containing full sentences and it is therefore well suited for creating linguistic resources such as Web corpora.
        The key idea is that long blocks can often be classified with high confidence, while shorter blocks require context-based adjustments.

        Here is an overview of the jusText algorithm:
            • Segmentation: The document is split into textual blocks based on HTML tags that typically define separate sections (e.g., <div>, <p>, <table>).
            • Preprocessing: Contents of <header>, <style>, and <script> tags are removed.
                Certain elements (e.g., <select>, copyright symbols) are immediately classified as boilerplate.
            • Context-Free Classification: Each block is classified as:
                - Bad (boilerplate) if it has high link density.
                - Short if it is too small to be classified reliably.
                - Near-Good if it has a moderate density of stopwords.
                - Good (main content) if it is long and contains many stopwords.
            • Context-Sensitive Classification: Blocks that were classified as short or near-good are reclassified based on surrounding blocks.
                The assumption is that main content clusters together, as does boilerplate.
            • Headings Processing: Header elements (e.g., <h1>, <h2>) are treated separately to ensure useful headings are preserved.
                Short headers near good content may be reclassified as near-good or good.

        Please refer to the jusText documentation for more details: https://corpus.tools/wiki/Justext/Algorithm

        Args:
            length_low: Minimum length of text to be considered for extraction.
            length_high: Maximum length of text to be considered for extraction.
            stopwords_low: Minimum proportion of stopwords in the text to be considered for extraction.
            stopwords_high: Maximum proportion of stopwords in the text to be considered for extraction.
            max_link_density: Maximum allowed link density in the text.
            max_heading_distance: Maximum distance from a heading to consider text for extraction.
            no_headings: If True, text extraction will ignore headings.
            is_boilerplate: If True, text extraction will ignore boilerplate content.
                Default is True for space-separated languages and False for non-space-separated languages
                (Thai, Chinese, Japanese, and Korean).
            logger: Optional logger instance for logging messages.

        """
        self.length_low = length_low
        self.length_high = length_high
        self.stopwords_low = stopwords_low
        self.stopwords_high = stopwords_high
        self.max_link_density = max_link_density
        self.max_heading_distance = max_heading_distance
        self.no_headings = no_headings
        self.is_boilerplate = is_boilerplate
        self.logger = logger

    def extract_text(self, html, stop_words, language):
        # Segment the HTML into paragraphs
        try:
            # Form the DOM tree
            dom = justext.core.html_to_dom(html)
            cleaned_dom = justext.core.preprocessor(dom)
            # Get the paragraphs from the DOM
            handler = justext.core.ParagraphMaker()
            lxml.sax.saxify(cleaned_dom, handler)
        except (lxml.etree.ParserError, ValueError, Exception):
            # Return nothing when we cannot segment the document
            if self.logger is not None:
                self.logger.info("Could not segment paragaphs in the document")
            return

        paragraphs = handler.paragraphs

        # Context free classification
        justext.core.classify_paragraphs(
            paragraphs,
            stop_words,
            self.length_low,
            self.length_high,
            self.stopwords_low,
            self.stopwords_high,
            self.max_link_density,
            self.no_headings,
        )

        # Copy the context free class to the class_style
        # This handles the headings as described in the
        # documentation
        for paragraph in paragraphs:
            paragraph.class_type = paragraph.cf_class

        # Context sensitive classification
        justext.core.revise_paragraph_classification(
            paragraphs,
            self.max_heading_distance,
        )

        if self.is_boilerplate is None:
            if language in NON_SPACED_LANGUAGES:
                warnings.warn("Disabling is_boilerplate check for jusText extraction.")
                is_boilerplate = False
            else:
                is_boilerplate = True

        else:
            is_boilerplate = self.is_boilerplate

        if is_boilerplate:
            return [p.text for p in paragraphs if not p.is_boilerplate]

        else:
            return [p.text for p in paragraphs]


class ResiliparseExtractor(HTMLExtractorAlgorithm):
    def __init__(
        self,
        required_stopword_density=0.32,
        main_content=True,
        alt_texts=False,
    ):
        """
        Initialize the Resiliparse text extraction algorithm with specified parameters.

        The Resiliparse algorithm extracts structural or semantic information from noisy raw web data for further processing,
        such as (main) content extraction / boilerplate removal, schema extraction, general web data cleansing, and more.

        It is implemented via the `extract_plain_text` function in the `resiliparse.extract.html2text` module.
        Resiliparse HTML2Text is a very fast and rule-based plain text extractor for HTML pages which uses the Resiliparse DOM parser.
        The `extract_plain_text` function extracts all visible text nodes inside the HTML document's <body>.
        Only <script>, <style> and a few other (generally) invisible elements are skipped and very basic ASCII formatting is applied.

        Please refer to the Resiliparse documentation for more details: https://resiliparse.chatnoir.eu/en/latest/man/extract/html2text.html

        NeMo Curator has added a stopword density filter to the Resiliparse extraction process, which requires that a paragraph contains a certain proportion of stopwords.

        Args:
            required_stopword_density: Proportion of stopwords required preserve an extracted paragraph.
                Studies on stopword lists and their distribution in various text corpora often
                suggest that around 30-40% of a typical English text consists of stopwords.
            main_content: Whether to apply simple heuristics for extracting only "main-content" elements.
            alt_texts: Whether to preserve alternative text descriptions (e.g., for images).

        """
        self.required_stopword_density = required_stopword_density
        self.main_content = main_content
        self.alt_texts = alt_texts

    def extract_text(self, html, stop_words, language):
        text = extract_plain_text(
            html, main_content=self.main_content, alt_texts=self.alt_texts
        )

        paragraphs = list(filter(None, text.split("\n")))

        if language in NON_SPACED_LANGUAGES:
            warnings.warn(
                "stopword_density is ignored for non-space-separated languages."
            )
            result = paragraphs
        else:
            result = []

            for paragraph in paragraphs:
                words = paragraph.split()
                length = len(words)

                if length == 0:
                    continue

                stopwords = [word for word in words if word in stop_words]
                stopword_density = len(stopwords) / length

                if stopword_density >= self.required_stopword_density:
                    result.append(paragraph)

        return result


class TrafilaturaExtractor(HTMLExtractorAlgorithm):
    def __init__(
        self,
        required_stopword_density=0.32,
        min_extracted_size=250,
        min_extracted_comm_size=1,
        min_output_size=1,
        min_output_comm_size=1,
        max_tree_size=None,
        min_duplcheck_size=100,
        max_repetitions=2,
        **extract_kwargs,
    ):
        """
        Initialize the Trafilatura text extraction algorithm with specified parameters.

        The Trafilatura extraction process combines readability-lxml and jusText as fallbacks to ensure robustness.
        Trafilatura's own algorithm follows a cascade of rule-based filters and content heuristics:
            • Content Delimitation: Uses XPath expressions to exclude unwanted HTML elements (e.g., navigation bars) and focus on relevant content (e.g., article body).
                Extracted HTML nodes are analyzed for relevance based on element type, text length, and link density.
            • Fallback Mechanism: If extraction seems faulty, alternative algorithms are run as backups.
                These use heuristics like line length, text-to-markup ratio, and HTML depth to improve extraction.
                Outputs are compared, prioritizing longer extractions with fewer impurities.
            • Baseline Extraction: If all else fails, it searches for text elements that might have been missed, discarding irrelevant content.

        The system balances precision and recall, extracting main text, comments, and metadata (title, site name, author, date, categories, tags).

        Please refer to the Trafilatura documentation for more details:
            https://trafilatura.readthedocs.io/en/latest/ and https://aclanthology.org/2021.acl-demo.15/

        NeMo Curator has added a stopword density filter to the Trafilatura extraction process, which requires that a paragraph contains a certain proportion of stopwords.

        Args:
            required_stopword_density: Proportion of stopwords required preserve an extracted paragraph.
                Studies on stopword lists and their distribution in various text corpora often
                suggest that around 30-40% of a typical English text consists of stopwords.
            min_extracted_size: Acceptable size in characters (used to trigger fallbacks).
                Defaults to 250. See Trafilatura documentation: https://trafilatura.readthedocs.io/en/latest/settings.html.
            min_extracted_comm_size: Works the same as min_output_comm_size for comment extraction.
                Defaults to 1. See Trafilatura documentation: https://trafilatura.readthedocs.io/en/latest/settings.html.
            min_output_size: Absolute acceptable minimum for main text output.
                Defaults to 1. See Trafilatura documentation: https://trafilatura.readthedocs.io/en/latest/settings.html.
            min_output_comm_size: Works the same as min_output_comm_size for comment extraction.
                Defaults to 1. See Trafilatura documentation: https://trafilatura.readthedocs.io/en/latest/settings.html.
            max_tree_size: Used to discard documents with too many elements. Defaults to None.
            min_duplcheck_size: Minimum size in characters to run deduplication on.
                Defaults to 100. See Trafilatura documentation: https://trafilatura.readthedocs.io/en/latest/settings.html.
            max_repetitions: Maximum number of duplicates allowed.
                Defaults to 2. See Trafilatura documentation: https://trafilatura.readthedocs.io/en/latest/settings.html.
            extract_kwargs: Additional keyword arguments for the Trafilatura extract function.
                See API documentation https://trafilatura.readthedocs.io/en/latest/corefunctions.html#extract
                for list of possible parameters.
                All arguments are set to their default values, except for deduplicate (bool) which is set to True.

        """
        self.required_stopword_density = required_stopword_density
        self.min_extracted_size = min_extracted_size
        self.min_extracted_comm_size = min_extracted_comm_size
        self.min_output_size = min_output_size
        self.min_output_comm_size = min_output_comm_size
        self.max_tree_size = max_tree_size
        self.min_duplcheck_size = min_duplcheck_size
        self.max_repetitions = max_repetitions
        self.extract_kwargs = extract_kwargs

    def extract_text(self, html, stop_words, language):
        trafilatura_config = deepcopy(TRAFILATURA_DEFAULT_CONFIG)
        trafilatura_config["DEFAULT"]["MIN_EXTRACTED_SIZE"] = str(
            self.min_extracted_size
        )
        trafilatura_config["DEFAULT"]["MIN_EXTRACTED_COMM_SIZE"] = str(
            self.min_extracted_comm_size
        )
        trafilatura_config["DEFAULT"]["MIN_OUTPUT_SIZE"] = str(self.min_output_size)
        trafilatura_config["DEFAULT"]["MIN_OUTPUT_COMM_SIZE"] = str(
            self.min_output_comm_size
        )
        if self.max_tree_size:
            trafilatura_config["DEFAULT"]["MAX_TREE_SIZE"] = str(self.max_tree_size)
        trafilatura_config["DEFAULT"]["MIN_DUPLCHECK_SIZE"] = str(
            self.min_duplcheck_size
        )
        trafilatura_config["DEFAULT"]["MAX_REPETITIONS"] = str(self.max_repetitions)

        # Recommended to set deduplicate=True
        self.extract_kwargs.setdefault("deduplicate", True)

        text = extract_with_trafilatura(
            html, config=trafilatura_config, **self.extract_kwargs
        )

        if text is not None:
            paragraphs = list(filter(None, text.split("\n")))

            if language in NON_SPACED_LANGUAGES:
                warnings.warn(
                    "stopword_density is ignored for non-space-separated languages."
                )
                result = paragraphs

            else:
                result = []

                for paragraph in paragraphs:
                    words = paragraph.split()
                    length = len(words)

                    if length == 0:
                        continue

                    stopwords = [word for word in words if word in stop_words]
                    stopword_density = len(stopwords) / length

                    if stopword_density >= self.required_stopword_density:
                        result.append(paragraph)

        else:
            return None

        if len(result) == 0:
            return None
        return result


def get_stop_list_dict(languages=[]):

    # Name mapping for language names from CLD2 (values)
    # and jusText (keys)
    lang_map = {
        "Haitian": "HAITIAN_CREOLE",
        "Norwegian_Bokmal": "NORWEGIAN",
        "Norwegian_Nynorsk": "NORWEGIAN_N",
        "Waray_Waray": "WARAY_PHILIPPINES",
    }

    # List obtained from https://github.com/stopwords-iso/stopwords-ja
    from .ja_stopwords import ja_stopwords

    # List obtained from https://github.com/stopwords-iso/stopwords-th
    from .th_stopwords import th_stopwords

    # List obtained from https://github.com/stopwords-iso/stopwords-zh
    from .zh_stopwords import zh_stopwords

    custom_stopwords = {
        "THAI": th_stopwords,
        "CHINESE": zh_stopwords,
        "JAPANESE": ja_stopwords,
    }

    if len(languages) == 0:
        languages = justext.get_stoplists()

        # Remove Latin as it yields a lot of low quality documents
        languages = list(languages)
        languages.remove("Latin")

        # Manually add Thai, Chinese, and Japanese
        languages.append("THAI")
        languages.append("CHINESE")
        languages.append("JAPANESE")

        languages = frozenset(languages)

    stop_list_dict = {}
    for language in languages:
        if language in lang_map:
            lang_key = lang_map[language]
        else:
            lang_key = language.upper()

        if lang_key in custom_stopwords:
            stop_list_dict[lang_key] = custom_stopwords[lang_key]
        else:
            stop_list_dict[lang_key] = justext.get_stoplist(language)

    return stop_list_dict


def get_all_stop_words():
    stop_words = set()
    for language in justext.get_stoplists():
        stop_words.update(justext.get_stoplist(language))

    return frozenset(stop_words)


class CommonCrawlWARCDownloader(DocumentDownloader):
    """
    Downloads WARC files from the Common Crawl
    """

    def __init__(self, download_dir, aws=False, verbose=False):
        """
        Creates a downloader

        Args:
          download_dir: Path to store raw compressed WARC files
          aws: If True, uses the s5cmd command to download from the Common Crawl's S3 bucket.
            If False, uses wget.
          verbose: If True, logs stdout and stderr of the download command (s5cmd/wget)
        """
        super().__init__()
        self._download_dir = download_dir
        self._aws = aws
        self._verbose = verbose

    def download(self, url):
        # Download each URL to the directory
        urlpath = urlparse(url).path[1:]
        output_name = urlpath.replace("/", "-")
        output_file = os.path.join(self._download_dir, output_name)
        if os.path.exists(output_file):
            print(f"WARC file: {output_file} exists. Not downloading")
        else:
            print(f"Downloading {url} and writing to {output_file}")
            # Download with either wget or s5cmd (aws)
            if self._aws:
                s3path = os.path.join("s3://commoncrawl/", urlpath)
                cmd = ["s5cmd", "cp", s3path, output_file]
            else:
                cmd = ["wget", url, "-O", output_file]
            if self._verbose:
                stdout, stderr = None, None
            else:
                stdout, stderr = subprocess.DEVNULL, subprocess.DEVNULL
            p = subprocess.run(
                cmd,
                stdout=stdout,
                stderr=stderr,
            )
            if p.returncode != 0:
                print(f"Failed to download {url} to {output_file}")

        return output_file


class CommonCrawlWARCDownloaderExtractOnly(DocumentDownloader):
    """
    A 'dummy' downloader that simply puts pre-downloaded
    files on the queue
    """

    def __init__(self, aws=False, verbose=False):
        super().__init__()

    def download(self, url):
        print(f"Putting WARC file {url} on the queue for extraction")
        return url


class CommonCrawlWARCIterator(DocumentIterator):

    def __init__(self, log_frequency=1000):
        super().__init__()
        self._counter = 0
        self._log_frequency = log_frequency

    def iterate(self, file_path):
        # Loop over all records in the current WARC
        self._counter = 0
        bname = os.path.split(file_path)[-1]
        with open(file_path, "rb") as file_pointer:
            ai = ArchiveIterator(file_pointer, arc2warc=True)
            for k, rec in enumerate(ai):
                # Get the response from the crawl
                if rec.rec_type == "response":
                    if self._counter > 0 and self._counter % self._log_frequency == 0:
                        print(f"Extracted {self._counter} records in WARC")
                    self._counter += 1
                    content = rec.content_stream().read()
                    warc_id = rec.rec_headers.get_header("WARC-Record-ID")[10:-1]
                    url = rec.rec_headers.get_header("WARC-Target-URI")
                    meta = {
                        "url": url,
                        "warc_id": warc_id,
                        "source_id": f"{bname}",
                    }
                    yield meta, content


class CommonCrawlWARCExtractor(DocumentExtractor):

    def __init__(self, algorithm=JusTextExtractor(), stop_lists=None):
        if stop_lists is not None:
            self._stop_lists = stop_lists
        else:
            self._stop_lists = get_stop_list_dict()

        self.algorithm = algorithm
        super().__init__()

    def extract(self, content):
        html = decode_html(content)
        if html is not None:
            # Language detection and HTML extraction
            lang = lang_detect(html)
            text = None
            if lang in self._stop_lists:
                text = self.algorithm.extract_text(html, self._stop_lists[lang], lang)
            if text is not None:
                if len(text) > 0:
                    text = "\n\n".join(text)
                    meta = {"language": lang, "text": text}
                    return meta
                else:
                    return None


def download_common_crawl(
    output_path: str,
    start_snapshot: str,
    end_snapshot: str,
    output_type: Literal["jsonl", "parquet"] = "jsonl",
    algorithm=JusTextExtractor(),
    stop_lists=None,
    news: bool = False,
    aws: bool = False,
    raw_download_dir: Optional[str] = None,
    keep_raw_download: bool = False,
    force_download: bool = False,
    url_limit: Optional[int] = None,
    record_limit: Optional[int] = None,
) -> DocumentDataset:
    """
    Downloads Common Crawl WARC snapshots and extracts text content using a specified extraction algorithm.

    Args:
      output_path (str): The root directory used for managing download and extraction.
          • Raw WARC files are stored in a "downloads" subdirectory under this path.
          • This path is also checked for existing extraction results; if found, extraction can be skipped.
          • Note: This function returns a DocumentDataset, and writing the extracted data to disk is the caller's responsibility.
      start_snapshot (str): Identifier for the earliest snapshot to process.
          • For CC-MAIN datasets, use the format 'YYYY-WeekNumber' (e.g., '2020-50' or '2021-04').
          • For CC-NEWS datasets (when news=True), use the 'YYYY-MM' (Year-Month) format.
      end_snapshot (str): Identifier for the latest snapshot to process, which must be chronologically after start_snapshot.
      output_type (Literal["jsonl", "parquet"]): The file format for the extracted output. Must be either "jsonl" or "parquet".
          • This is not used for the output file, but is used to check if an extracted output already exists.
      algorithm: The text extraction algorithm instance to use for HTML processing.
          • This can be a JusTextExtractor (default), ResiliparseExtractor, or TrafilaturaExtractor object.
      stop_lists: A dictionary stop lists, where the keys are languages (e.g., "ENGLISH")
          and the values are Python frozensets denoting the list of stop words for that language.
          If None, it defaults to jusText's stop lists: https://github.com/miso-belica/jusText/tree/main/justext/stoplists,
          with added Thai, Chinese, and Japanese support.
      news (bool): When True, indicates that URLs should be retrieved from the CC-NEWS dataset.
          • This also means snapshot identifiers should follow the 'YYYY-MM' format.
      aws (bool): If True, downloads are sourced from Common Crawl's S3 bucket using s5cmd;
          • If False, wget is used to fetch the files via HTTPS.
      raw_download_dir: Optional; the directory to temporarily store raw WARC files.
          • If not provided, defaults to a "downloads" folder within output_path.
      keep_raw_download (bool): If True, retains the downloaded raw WARC files after extraction.
          • If False, these raw files may be removed following extraction.
      force_download (bool): If False, skips re-downloading or re-extracting snapshots if outputs already exist in output_path.
      url_limit: Optional; the maximum number of WARC files to download from the snapshot range.
          • If None, all available files within the specified snapshots are downloaded.
      record_limit: Optional; the maximum number of records to extract from each WARC file.
          • If None, all available records are extracted.
    """
    common_crawl_urls = get_common_crawl_urls(
        starting_snapshot=start_snapshot, ending_snapshot=end_snapshot, news=news
    )

    if len(common_crawl_urls) == 0:
        raise ValueError(
            f"No Common Crawl download urls found between {start_snapshot} and {end_snapshot}. "
            "Ensure that a valid Common Crawl snapshot (https://data.commoncrawl.org/) is "
            "within the range provided."
        )

    if url_limit:
        common_crawl_urls = common_crawl_urls[:url_limit]
    output_paths = list(
        map(
            lambda url: os.path.join(
                output_path, url.split("/")[-1] + f".{output_type}"
            ),
            common_crawl_urls,
        )
    )

    if not raw_download_dir:
        raw_download_dir = os.path.join(output_path, "downloads")
    expand_outdir_and_mkdir(raw_download_dir)
    downloader = CommonCrawlWARCDownloader(raw_download_dir, aws=aws)
    iterator = CommonCrawlWARCIterator()
    extractor = CommonCrawlWARCExtractor(algorithm=algorithm, stop_lists=stop_lists)

    output_format = {
        "text": str,
        "language": str,
        "url": str,
        "warc_id": str,
        "source_id": str,
        "file_name": str,
    }
    dataset = download_and_extract(
        common_crawl_urls,
        output_paths,
        downloader,
        iterator,
        extractor,
        output_format,
        output_type=output_type,
        keep_raw_download=keep_raw_download,
        force_download=force_download,
        filename_col="file_name",
        record_limit=record_limit,
    )

    return dataset



# Copyright (c) 2024, NVIDIA CORPORATION.  All rights reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import dask.dataframe as dd
import pandas as pd
import pytest

import nemo_curator as nc
from nemo_curator.datasets import DocumentDataset
from nemo_curator.utils.import_utils import gpu_only_import, is_unavailable

cudf = gpu_only_import("cudf")
is_cudf_available = not is_unavailable(cudf)


def list_to_dataset(documents, col_name="text", npartitions=2, backend="pandas"):
    data = {col_name: documents}
    pdf = pd.DataFrame(data)
    ddf = dd.from_pandas(pdf, npartitions=npartitions)
    if backend == "cudf" and is_unavailable(cudf):
        raise ImportError("cuDF is not installed or importable.")
    ddf = ddf.to_backend(backend)
    return DocumentDataset(ddf)


@pytest.fixture(params=["pandas", pytest.param("cudf", marks=pytest.mark.gpu)])
def single_partition_dataset(request):
    return list_to_dataset(
        ["First", "Second", "Third", "Fourth", "Fifth"],
        npartitions=1,
        backend=request.param,
    )


@pytest.fixture(params=["pandas", pytest.param("cudf", marks=pytest.mark.gpu)])
def two_partition_dataset(request):
    return list_to_dataset(
        ["First", "Second", "Third", "Fourth", "Fifth"],
        npartitions=2,
        backend=request.param,
    )


class TestAddId:
    def test_basic_id(self, single_partition_dataset):
        id_field = "id"
        add_id = nc.AddId(id_field, start_index=0)
        id_dataset = add_id(single_partition_dataset)
        actual_ids = id_dataset.df[id_field].compute()
        expected_ids = pd.Series(
            [
                "doc_id-0000000000",
                "doc_id-0000000001",
                "doc_id-0000000002",
                "doc_id-0000000003",
                "doc_id-0000000004",
            ]
        )
        if is_cudf_available and isinstance(actual_ids, cudf.Series):
            actual_ids = actual_ids.to_pandas()

        assert all(
            expected_ids == actual_ids
        ), f"Expected: {expected_ids}, got: {actual_ids}"

    def test_two_partitions(self, two_partition_dataset):
        id_field = "id"
        add_id = nc.AddId(id_field, start_index=0)
        id_dataset = add_id(two_partition_dataset)
        actual_ids = id_dataset.df[id_field].compute()
        expected_ids = pd.Series(
            [
                "doc_id-0000000000",
                "doc_id-0000000001",
                "doc_id-0000000002",
                "doc_id-0000000003",
                "doc_id-0000000004",
            ]
        )
        if is_cudf_available and isinstance(actual_ids, cudf.Series):
            actual_ids = actual_ids.to_pandas()

        assert all(
            expected_ids == actual_ids
        ), f"Expected: {expected_ids}, got: {actual_ids}"

    def test_id_prefix(self, two_partition_dataset):
        id_field = "id"
        id_prefix = "my_id"
        add_id = nc.AddId(id_field, id_prefix=id_prefix, start_index=0)
        id_dataset = add_id(two_partition_dataset)
        actual_ids = id_dataset.df[id_field].compute()
        expected_ids = pd.Series(
            [
                f"{id_prefix}-0000000000",
                f"{id_prefix}-0000000001",
                f"{id_prefix}-0000000002",
                f"{id_prefix}-0000000003",
                f"{id_prefix}-0000000004",
            ]
        )
        if is_cudf_available and isinstance(actual_ids, cudf.Series):
            actual_ids = actual_ids.to_pandas()

        assert all(
            expected_ids == actual_ids
        ), f"Expected: {expected_ids}, got: {actual_ids}"

    def test_start_index(self, two_partition_dataset):
        id_field = "id"
        start_index = 13
        add_id = nc.AddId(id_field, start_index=start_index)
        id_dataset = add_id(two_partition_dataset)
        actual_ids = id_dataset.df[id_field].compute()
        expected_ids = pd.Series(
            [
                "doc_id-0000000013",
                "doc_id-0000000014",
                "doc_id-0000000015",
                "doc_id-0000000016",
                "doc_id-0000000017",
            ]
        )
        if is_cudf_available and isinstance(actual_ids, cudf.Series):
            actual_ids = actual_ids.to_pandas()

        assert all(
            expected_ids == actual_ids
        ), f"Expected: {expected_ids}, got: {actual_ids}"

    def test_fast_id_single_partition(self, single_partition_dataset):
        id_field = "id"
        add_id = nc.AddId(id_field)
        id_dataset = add_id(single_partition_dataset)
        actual_ids = id_dataset.df[id_field].compute()
        expected_ids = pd.Series(
            [
                "doc_id-00",
                "doc_id-10",
                "doc_id-20",
                "doc_id-30",
                "doc_id-40",
            ]
        )
        if is_cudf_available and isinstance(actual_ids, cudf.Series):
            actual_ids = actual_ids.to_pandas()

        assert all(
            expected_ids == actual_ids
        ), f"Expected: {expected_ids}, got: {actual_ids}"

    def test_fast_id_two_partitions(self, two_partition_dataset):
        id_field = "id"
        add_id = nc.AddId(id_field)
        id_dataset = add_id(two_partition_dataset)
        actual_ids = id_dataset.df[id_field].compute()
        expected_ids = pd.Series(
            [
                "doc_id-00",
                "doc_id-10",
                "doc_id-20",
                "doc_id-01",
                "doc_id-11",
            ]
        )
        if is_cudf_available and isinstance(actual_ids, cudf.Series):
            actual_ids = actual_ids.to_pandas()

        assert all(
            expected_ids == actual_ids
        ), f"Expected: {expected_ids}, got: {actual_ids}"



# Copyright (c) 2024, NVIDIA CORPORATION.  All rights reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import argparse
import csv

from nemo_curator.datasets import DocumentDataset
from nemo_curator.modifiers import FastTextLabelModifier
from nemo_curator.modules import Modify
from nemo_curator.utils.distributed_utils import get_client, read_data
from nemo_curator.utils.file_utils import get_all_files_paths_under
from nemo_curator.utils.script_utils import ArgumentHelper


def sample_rows(df, n, seed):
    samples = df.sample(frac=n / len(df) + 0.05, random_state=seed)

    return samples.head(n=n, compute=False)


def main(args):
    client = get_client(**ArgumentHelper.parse_client_args(args))
    # Get local path
    files = list(
        get_all_files_paths_under(args.input_data_dir, keep_extensions="jsonl")
    )
    raw_data = read_data(files, file_type="jsonl", backend="pandas")
    dataset = DocumentDataset(raw_data)
    text_field = args.input_json_field

    # fastText requires each document to be prepended with a special label for training
    preprocessing = Modify(FastTextLabelModifier(args.label), text_field=text_field)
    labeled_data = preprocessing(dataset)

    samples = sample_rows(labeled_data.df, args.output_num_samples, args.seed)

    samples[text_field].to_csv(
        args.output_train_file,
        single_file=True,
        encoding="utf-8",
        header=False,
        index=False,
        quoting=csv.QUOTE_NONE,
        sep="\n",
    )
    client.close()


def attach_args(
    parser=argparse.ArgumentParser(
        """
Prepare data for training skip-gram classifier with FastText.

Takes as input a directory of .jsonl files, and writes an output
file of samples prepared in order to train a skip-gram classifier
with FastText.
""",
        formatter_class=argparse.ArgumentDefaultsHelpFormatter,
    )
):
    argumentHelper = ArgumentHelper(parser)

    argumentHelper.add_arg_input_data_dir()
    argumentHelper.add_arg_log_dir(default="./log/prepare_filter_data")
    argumentHelper.add_arg_output_train_file(
        help="The output file containing prepared samples to train a "
        "skip-gram classifier with FastText."
    )
    argumentHelper.add_arg_seed(
        help="The random seed to use for sampling from the dataset."
    )
    argumentHelper.add_distributed_args()
    parser.add_argument(
        "--input-json-field",
        type=str,
        default="text",
        help="The input field within each JSON object on which the filter will "
        'operate. By default, the filter will operate on the "text" '
        'field, but other fields can be specified, such as "url" or "id".',
    )
    parser.add_argument(
        "--label",
        type=str,
        default=None,
        required=True,
        help="The label to be used at the beginning of each sample "
        'in the output file. For example, "__label__hq" could be '
        "used for the high-quality (positive) samples.",
    )

    return parser


def console_script():
    main(attach_args().parse_args())



# Copyright (c) 2024, NVIDIA CORPORATION.  All rights reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from __future__ import annotations

from typing import Tuple

import numba
import numpy as np

from nemo_curator._compat import DASK_SHUFFLE_METHOD_ARG, query_planning_enabled


def get_agg_text_bytes_df(
    df: dask_cudf.DataFrame,
    agg_column: str,
    bytes_column: str,
    n_partitions: int,
    shuffle: bool = False,
) -> Tuple[dask_cudf.DataFrame, int]:
    """
    Groupby bucket and calculate total bytes for a bucket.
    """
    if query_planning_enabled():
        # `shuffle_method: bool` doesn't really make sense
        # when query-planning is enabled, because dask-expr
        # will ALWAYS use a shuffle-based reduction when
        # `split_out>1`
        shuffle_arg = {}
    else:
        shuffle_arg = {
            "shuffle_method" if DASK_SHUFFLE_METHOD_ARG else "shuffle": shuffle
        }
    agg_df = (
        df[[agg_column, bytes_column]]
        .groupby([agg_column])
        .agg({bytes_column: "sum"}, split_out=n_partitions, **shuffle_arg)
    )
    agg_df = agg_df.reset_index(drop=False)
    # Doing a per partition sort
    # seems to cause issues with
    # jaccard shuffle  (Overflow errors)
    # which  are caught and then
    # retried with text_bytes_aware_merge
    agg_df = agg_df.persist()
    agg_df = agg_df.sort_values(by=[bytes_column], ascending=False, ignore_index=True)
    agg_df = agg_df.persist()
    # Added length to force computation
    # after persist
    agg_df_len = len(agg_df)

    return agg_df, agg_df_len


# next-fit-descending bin packing
# https://en.wikipedia.org/wiki/Next-fit-decreasing_bin_packing
@numba.jit(nopython=True)
def build_partition(sizes: np.ndarray, max_size: int) -> np.ndarray:
    """
    Given an array of items and a max bin size this method
    attempts to return a grouping of items such that no group exceeds
    the max bin size using the Next-fit-decreasing bin packing approach.
    """
    i: int = 0
    count: int = 0
    current: int = 0
    size: int = 0
    partition = np.empty(sizes.shape, dtype=np.int32)
    for i in range(len(sizes)):
        size = sizes[i]
        if current + size < max_size:
            partition[i] = count
            current += size
        else:
            count += 1
            current = size
            partition[i] = count
    return partition



# Copyright (c) 2024, NVIDIA CORPORATION.  All rights reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import hashlib
import importlib
import re
import secrets
from typing import Any

import pandas as pd
from dask.base import normalize_token
from openai import OpenAI
from tqdm import tqdm

from nemo_curator import OpenAIClient
from nemo_curator.datasets import DocumentDataset
from nemo_curator.synthetic import NemotronGenerator
from nemo_curator.synthetic.generator import SyntheticDataGenerator
from nemo_curator.utils.distributed_utils import load_object_on_worker

tqdm.pandas()
config = importlib.import_module(
    "tutorials.nemo-retriever-synthetic-data-generation.config.config"
)
RetrieverEvalSDGConfig = config.RetrieverEvalSDGConfig


# ----------------------------------------------------------------------------80
class RetrieverEvalSetGenerator(SyntheticDataGenerator):
    """
    Main class that generates annotated datasets for retriever evaluation
    Question, Answers are generated for a given document chunk as input
    Datasets are annotated in format of (passage, question, answer) triplets
    """

    def __init__(
        self,
        pipeline_config: RetrieverEvalSDGConfig = None,
    ):
        super().__init__()
        self._name = self.__class__.__name__
        self.cfg = pipeline_config

        self._init_pipeline_params()

    def load_pipeline_config(self, cfg_path: str):
        self.cfg = RetrieverEvalSDGConfig.from_yaml(cfg_path)

    def _validate_config(self):
        return True  # TODO complete this

    def _init_pipeline_params(self):

        if self._validate_config():
            self.sys_prompt = self.cfg.generator_system_prompt
            self.user_prompt_template = self.cfg.generator_user_prompt_template
            self.generator_model = self.cfg.generator_model
            self.generator_model_kwargs = {
                "temperature": self.cfg.temperature,
                "top_p": self.cfg.top_p,
                "max_tokens": self.cfg.max_tokens,
            }
            self.num_qs = self.cfg.num_questions
        else:
            raise Exception("Validation Error: incorrect pipeline config file")

    # ----------------------------------------------------------------------------80

    def _create_generator(self):
        openai_client = OpenAI(
            base_url=self.cfg.base_url,
            api_key=self.cfg.api_key,
        )
        client = OpenAIClient(openai_client)
        generator = NemotronGenerator(client)
        return generator

    def _get_partition_id(self, df: pd.DataFrame, partition_info=None):
        df["partition-id"] = partition_info["number"]
        return df

    def __call__(self, dataset: DocumentDataset) -> DocumentDataset:

        ddf = dataset.df
        ddf["partition-id"] = ""
        ddf = ddf.map_partitions(self._get_partition_id, meta=ddf)
        ddf["llm_response"] = ""
        ddf["qa_pairs"] = ""
        ddf["question"] = ""
        ddf["answer"] = ""
        if "_id" not in ddf.columns:
            ddf["_id"] = ""
        ddf["question-id"] = ""
        ddf["score"] = ""

        ddf = ddf.map_partitions(self._process_on_partition, meta=ddf)

        return DocumentDataset(ddf)

    def _process_on_partition(self, df: pd.DataFrame) -> pd.DataFrame:

        self.generator = load_object_on_worker(
            attr="generator",
            load_object_function=self._create_generator,
            load_object_kwargs={},
        )

        _id = df["partition-id"].iloc[0]
        tqdm.pandas(desc=f"For partition_{_id}")

        if "_id" in df.columns:
            df["_id"] = df["_id"].apply(self._check_doc_id)
        else:
            df["_id"] = df["text"].apply(self._get_random_hash)

        df["llm_response"] = df["text"].progress_apply(self.generate)
        df["qa_pairs"] = df["llm_response"].apply(self.parse_response)

        df = df.explode("qa_pairs").reset_index(drop=True)
        df["question"] = df["qa_pairs"].apply(lambda x: x["question"])
        df["question-id"] = df["question"].apply(self._get_random_hash)
        df["answer"] = df["qa_pairs"].apply(lambda x: x["answer"])
        df["score"] = df["question"].apply(lambda x: 1)

        return df

    # ----------------------------------------------------------------------------80
    def parse_response(self, llm_response: str) -> Any:
        qa_pairs = []
        if llm_response:
            qa_list = llm_response.split("Question")[1:]
            try:
                for qa in qa_list:
                    qas = qa.split("Answer")
                    q = qas[0].split(":")[1].strip()
                    if re.search("Explanation", qas[1]):
                        a = qas[1].split("Explanation")[0].split(":")[1].strip()
                        explanation = qas[1].split("Explanation")[1].strip()
                    else:
                        a = qas[1].split(":")[1].strip()
                    qa_pairs.append({"question": q, "answer": a})
            except Exception as e:
                qa_pairs = [{"question": "", "answer": ""}]
                print(f"error: {e}")
        else:
            qa_pairs = [{"question": "", "answer": ""}]
        return qa_pairs

    # ----------------------------------------------------------------------------80
    def generate(self, doc_text):
        try:
            response = self.generator.generate_closed_qa_instructions(
                document=doc_text,
                prompt_template=self.sys_prompt + "\n" + self.user_prompt_template,
                n_openlines=self.num_qs,
                model=self.generator_model,
                model_kwargs=self.generator_model_kwargs,
            )
        except Exception as e:
            print(f"error: {e}")
            return ""

        return response[0]

    # ----------------------------------------------------------------------------80
    def _get_random_hash(self, question: str):
        """Generate random hash for synthetic question IDs"""
        # Generate a random string
        random_string = secrets.token_hex(16)
        # Generates a secure, random string of 16 bytes hex-encoded

        # Hash the random string using SHA-256
        hash_object = hashlib.sha256(
            random_string.encode()
        )  # Encode the string to bytes
        hex_dig = hash_object.hexdigest()
        return hex_dig

    # ----------------------------------------------------------------------------80
    def _check_doc_id(self, doc_id: Any) -> str:
        if doc_id:
            if str(doc_id) != "nan":
                return str(doc_id)
        return self._get_random_hash("some text")

    def __dask_tokenize__(self):
        return normalize_token(RetrieverEvalSetGenerator)


# ----------------------------------------------------------------------------80



# Copyright (c) 2025, NVIDIA CORPORATION.  All rights reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from __future__ import annotations

import os
import time
import warnings
from contextlib import nullcontext
from hashlib import md5
from typing import Optional, Union

import pandas as pd
from dask import config
from dask import dataframe as dd

from nemo_curator._compat import DASK_P2P_ERROR
from nemo_curator.datasets import DocumentDataset
from nemo_curator.log import create_logger
from nemo_curator.modules.base import BaseDeduplicationModule
from nemo_curator.utils.distributed_utils import performance_report_if_with_ts_suffix
from nemo_curator.utils.duplicates_removal import remove_duplicates
from nemo_curator.utils.gpu_utils import is_cudf_type


class ExactDuplicates(BaseDeduplicationModule):
    """Find exact duplicates in a document corpus"""

    SUPPORTED_HASHES = {"md5"}

    def __init__(
        self,
        logger: Union[logging.LoggerAdapter, str] = "./",
        id_field: str = "id",
        text_field: str = "text",
        hash_method: str = "md5",
        perform_removal: bool = False,
        profile_dir: Optional[str] = None,
        cache_dir: Optional[str] = None,
    ):
        """
        Parameters
        ----------
        logger: Existing logger to log to, or a path to a log directory.
        id_field: Column in the Dataset denoting document ID.
        text_field: Column in the Dataset denoting document content.
        hash_method: The hashing algorithm used for identifying exact duplicates. Currently supports {"md5"}
        profile_dir: str, Default None
          If specified directory to write dask profile
        cache_dir: str, Default None
          If specified, will compute & write duplicate id's to cache directory.
        """
        super().__init__(
            id_field=id_field,
            text_field=text_field,
            input_backend="any",
            logger=logger,
            perform_removal=perform_removal,
            profile_dir=profile_dir,
            cache_dir=cache_dir,
        )

        if hash_method not in self.SUPPORTED_HASHES:
            msg = f"{hash_method} not in supported hash_methods. Choose a hash_method from {self.SUPPORTED_HASHES}"
            raise ValueError(msg)

        self.hash_method = hash_method

        if isinstance(logger, str):
            self._logger = create_logger(
                rank=0,
                log_file=os.path.join(logger, "ExactDuplicates.log"),
                name="ExactDuplicates",
            )
        else:
            self._logger = logger

    def _exact_dup_ids(self, df: dd.DataFrame) -> dd.DataFrame:
        """
        Get the id's for text/documents that are exact duplicates
        Parameters
        ----------
        df: dask.dataframe.DataFrame
          A dataframe with the following requirements:
          * A column where each row is the text from one document
          * A unique ID column for each document
        """
        hash_df = self._compute_hashes(df)

        shuffle_context = (
            config.set({"dataframe.shuffle.method": "tasks"})
            if DASK_P2P_ERROR
            else nullcontext()
        )

        with shuffle_context:
            dup_ids = hash_df.shuffle(
                on=["_hashes"],
                ignore_index=True,
                npartitions=max(1, (hash_df.npartitions // 3)),
            ).map_partitions(lambda x: x[x["_hashes"].duplicated(keep=False)])

        return dup_ids

    def _compute_hashes(
        self,
        df: dd.DataFrame,
    ) -> dd.DataFrame:
        """
        Computes the hash of the text_column provided and returns a dataframe
        containing the id_column and relevant hashes in the _hashes column.
        """
        self._logger.info("Starting lazy hash generation")
        res = df[[self.id_field]]
        res["_hashes"] = df[self.text_field].map_partitions(self.hash_documents)

        self._logger.info(
            f"Lazy hash generation complete for {res.npartitions} partitions"
        )

        return res

    def hash_documents(
        self, df: Union[cudf.Series, pd.Series]
    ) -> Union[cudf.Series, pd.Series]:
        """
        Compute hashes for a Series containing documents
        """
        if is_cudf_type(df):
            return df.hash_values(method=self.hash_method)

        elif isinstance(df, pd.Series):
            # TODO: Generalize ty using self.hash_method
            return df.apply(lambda x: md5(x.encode()).hexdigest())

        else:
            raise ValueError(f"Unsupported type: {type(df)}")

    def identify_duplicates(self, dataset: DocumentDataset) -> DocumentDataset:
        """
        Find document IDs for exact duplicates in a given DocumentDataset
        Parameters
        ----------
        dataset: DocumentDataset
          The input datset to find exact duplicates
        Returns
        -------
        DocumentDataset containing IDs and hashes of all duplicate documents
        """
        result = self._exact_dup_ids(df=dataset.df)

        if self.cache_dir is None:
            return DocumentDataset(result)

        t0 = time.time()
        self._logger.info("Starting execution for ExactDedup")
        write_path = os.path.join(self.cache_dir, "_exact_duplicates.parquet")

        if os.path.exists(write_path):
            warnings.warn(
                f"Output path f{write_path} already exists and will be overwritten"
            )

        with performance_report_if_with_ts_suffix(
            self.profile_dir,
            "exact-dedup-profile",
        ):
            result.to_parquet(write_path, write_index=False, overwrite=True)

        self._logger.info(
            f"Time taken for Exact Dedup Computation = {time.time() - t0}s and output written at {write_path}"
        )

        backend = "cudf" if is_cudf_type(result) else "pandas"

        return DocumentDataset.read_parquet(
            write_path,
            backend=backend,
            # We read with files_per_partition=1 so that groups are read in whole (and do not exist across partitions)
            files_per_partition=1,
            blocksize=None,
        )

    def remove(
        self, dataset: DocumentDataset, duplicates_to_remove: Optional[DocumentDataset]
    ) -> DocumentDataset:
        """
        Remove exact duplicates from a given DocumentDataset
        Parameters
        ----------
        dataset: DocumentDataset
          The input dataset from which to remove exact duplicates
        duplicates_to_remove: DocumentDataset
          The dataset containing IDs of the exact duplicates to remove
        Returns
        -------
        DocumentDataset containing only non-duplicate documents
        """

        if duplicates_to_remove is None:
            print("No exact duplicates to remove, returning original dataset")
            return dataset

        result = remove_duplicates(
            left=dataset.df,
            duplicates=duplicates_to_remove.df,
            id_field=self.id_field,
            group_field="_hashes",
        )
        return DocumentDataset(result)



import csv
from typing import List, Tuple, Union

import dask.dataframe as dd
import pandas as pd

from nemo_curator.datasets.doc_dataset import DocumentDataset
from nemo_curator.utils.distributed_utils import _resolve_filename_col, write_to_disk
from nemo_curator.utils.file_utils import remove_path_extension
from nemo_curator.utils.import_utils import gpu_only_import

cudf = gpu_only_import("cudf")


class ParallelDataset(DocumentDataset):
    """
    An extension of the standard `DocumentDataset` with a special method that loads simple bitext.

    For data with more complicated metadata, please convert your data into jsonl/parquet/pickle format
    and use interfaces defined in `DocumentDataset`.
    """

    def persist(self):
        return ParallelDataset(self.df.persist())

    @classmethod
    def read_simple_bitext(
        cls,
        src_input_files: Union[str, List[str]],
        tgt_input_files: Union[str, List[str]],
        src_lang: str,
        tgt_lang: str,
        backend: str = "pandas",
        add_filename: Union[bool, str] = False,
        npartitions: int = 16,
    ):
        """See `read_single_simple_bitext_file_pair` docstring for what "simple_bitext" means and usage of other parameters.

        Args:
            src_input_files (Union[str, List[str]]): one or several input files, in source language
            tgt_input_files (Union[str, List[str]]): one or several input files, in target language

        Raises:
            TypeError: If types of `src_input_files` and `tgt_input_files` doesn't agree.

        Returns:
            ParallelDataset: A `ParallelDataset` object with `self.df` holding the ingested simple bitext.
        """

        if isinstance(src_input_files, str) and isinstance(tgt_input_files, str):
            src_input_files = [src_input_files]
            tgt_input_files = [tgt_input_files]
        elif not isinstance(src_input_files, list) or not isinstance(
            tgt_input_files, list
        ):
            raise TypeError("Both file inputs must be strings or lists.")

        # use default doc id for now
        # but in the future it might be useful to allow customizing doc id by passing a prefix
        df_files = []
        # We do not use `dd.from_map` because an individual file could be pretty large,
        # hence, it's not appropriate to partition based on individual files.
        # What we do is that we concatenate all the individual files and perform repartition.
        for src_input_file, tgt_input_file in zip(src_input_files, tgt_input_files):
            df_file = ParallelDataset.read_single_simple_bitext_file_pair(
                (src_input_file, tgt_input_file),
                src_lang=src_lang,
                tgt_lang=tgt_lang,
                backend=backend,
                add_filename=add_filename,
            )
            df_files.append(df_file)

        if backend == "cudf":
            df = cudf
        else:
            df = pd

        data = dd.from_pandas(df.concat(df_files), npartitions=npartitions)
        return cls(data)

    def to_bitext(
        self,
        output_file_dir,
        write_to_filename=False,
    ):
        """See `nemo_curator.utils.distributed_utils.write_to_disk` docstring for parameter usage."""
        write_to_disk(
            df=self.df,
            output_path=output_file_dir,
            write_to_filename=write_to_filename,
            output_type="bitext",
        )

    @staticmethod
    def read_single_simple_bitext_file_pair(
        input_file_pair: Tuple[str],
        src_lang: str,
        tgt_lang: str,
        doc_id: str = None,
        backend: str = "cudf",
        add_filename: Union[bool, str] = False,
    ) -> Union[dd.DataFrame, "dask_cudf.DataFrame"]:
        """This function reads a pair of "simple bitext" files into a pandas DataFrame.
        A simple bitext is a commonly data format in machine translation.
        It consists of two plain text files with the same number of lines, each line pair being translations of each other. For example:

        data.de:

        ```
        Wir besitzen keine Reisetaschen aus Leder.
        Die Firma produziert Computer für den deutschen Markt.
        ...
        ```

        data.en:

        ```
        We don't own duffel bags made of leather.
        The company produces computers for the German market.
        ...
        ```

        For simplicity, we also assume that the names of the two text files have the same prefix, except for different language code at the end as file extensions.

        Args:
            input_file_pair (Tuple[str]): A pair of file paths pointing to the input files
            src_lang (str): Source language, in ISO-639-1 (two character) format (e.g. 'en')
            tgt_lang (str): Target language, in ISO-639-1 (two character) format (e.g. 'en')
            doc_id (str, optional): A string document id to assign to every segment in the file. Defaults to None.
            backend (str, optional): Backend of the data frame. Defaults to "cudf".
            add_filename (Union[bool, str]): Whether to add a filename column to the DataFrame.
                If True, a new column is added to the DataFrame called `file_name`.
                If str, sets new column name. Default is False.


        Returns:
            Union[dd.DataFrame, dask_cudf.DataFrame]
        """
        src_input_file, tgt_input_file = input_file_pair
        assert remove_path_extension(src_input_file) == remove_path_extension(
            tgt_input_file
        ), f"Assuming source and target filenames would have common prefix before language code, but got {src_input_file} and {tgt_input_file}."

        if not doc_id:
            doc_id = "▁".join([src_input_file, tgt_input_file])

        if backend == "cudf":
            df = cudf
        else:
            df = pd

        df_src = df.read_csv(
            src_input_file, names=["src"], sep="\t", quoting=csv.QUOTE_NONE
        )
        df_tgt = df.read_csv(
            tgt_input_file, names=["tgt"], sep="\t", quoting=csv.QUOTE_NONE
        )
        assert len(df_src) == len(
            df_tgt
        ), f"We assume the source and target file would have the same number of lines, but got {len(df_src)} and {len(df_tgt)}."
        df_combined = df.concat([df_src, df_tgt], axis=1)
        df_combined["doc_id"] = doc_id
        df_combined["src_lang"] = src_lang
        df_combined["tgt_lang"] = tgt_lang

        if add_filename:
            df_combined[_resolve_filename_col(add_filename)] = remove_path_extension(
                src_input_file
            )

        return df_combined



# Copyright (c) 2024, NVIDIA CORPORATION.  All rights reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import logging
import os
import socket

from nemo_curator.utils.file_utils import expand_outdir_and_mkdir


def create_logger(rank, log_file, name="logger", log_level=logging.INFO, stdout=False):
    # Create the logger
    logger = logging.getLogger(name)
    logger.setLevel(log_level)

    myhost = socket.gethostname()

    extra = {"host": myhost, "rank": rank}
    formatter = logging.Formatter(
        "%(asctime)s | %(host)s | Rank %(rank)s | %(message)s"
    )

    # File handler for output
    file_handler = logging.FileHandler(log_file, mode="a")
    file_handler.setFormatter(formatter)
    logger.addHandler(file_handler)

    if stdout:
        stdout_handler = logging.StreamHandler()
        stdout_handler.setFormatter(formatter)
        logger.addHandler(stdout_handler)

    logger = logging.LoggerAdapter(logger, extra)
    return logger


def create_rank_logger(
    rank,
    log_dir,
    name="node_logger",
    log_level=logging.INFO,
):
    # Make the log directory if it does not exist
    log_dir = expand_outdir_and_mkdir(log_dir)

    # Create the rank subdirectory
    rank_tag = str(rank).rjust(3, "0")
    rank_dir = os.path.join(log_dir, f"rank_{rank_tag}")
    rank_dir = expand_outdir_and_mkdir(rank_dir)

    log_file = os.path.join(rank_dir, f"rank_{rank_tag}.log")
    return create_logger(rank, log_file, name=name, log_level=log_level)


def create_local_logger(
    rank,
    local_id,
    log_dir,
    name="local_logger",
    log_level=logging.INFO,
):
    # Create the logger
    logger = logging.getLogger(name)
    logger.setLevel(log_level)

    # Make tags
    rank_tag = str(rank).rjust(3, "0")
    local_id_tag = str(local_id).rjust(3, "0")

    myhost = socket.gethostname()
    extra = {"host": myhost, "node": rank_tag, "local": local_id_tag}
    formatter = logging.Formatter(
        "%(asctime)s | %(host)s | Node rank %(node)s "
        "| Local rank %(local)s | %(message)s"
    )

    # Output log file
    rank_dir = os.path.join(log_dir, f"rank_{rank_tag}")
    log_file = os.path.join(rank_dir, f"local_{local_id_tag}.log")

    # File handler for output
    file_handler = logging.FileHandler(log_file, mode="a")
    file_handler.setFormatter(formatter)
    logger.addHandler(file_handler)

    logger = logging.LoggerAdapter(logger, extra)

    return logger



import argparse
import logging
import os
import time

from nemo_curator.classifiers import QualityClassifier
from nemo_curator.datasets import DocumentDataset
from nemo_curator.utils.distributed_utils import get_client, get_num_workers
from nemo_curator.utils.file_utils import get_remaining_files

logging.basicConfig(format="%(asctime)s: %(message)s", level=logging.INFO)


SCHEDULER_FILE = os.environ.get("SCHEDULER_FILE")


if __name__ == "__main__":
    parser = argparse.ArgumentParser(description="Run NeMo quality classifier.")
    parser.add_argument("--input", help="Path to the input folder")
    parser.add_argument("--output", help="Path to the output folder")
    parser.add_argument(
        "--batch-size", type=int, default=64, help="Batch size for quality model"
    )
    args = parser.parse_args()

    t0 = time.time()
    client = get_client(scheduler_file=SCHEDULER_FILE)
    logging.info(f"Number of dask workers: {get_num_workers(client)}")
    logging.info(client)

    classifier = QualityClassifier(batch_size=args.batch_size)

    raw_base_path = args.input
    qm_base_path = args.output
    files = get_remaining_files(raw_base_path, qm_base_path, "parquet")
    if files:
        logging.info(f"Found {len(files)} remaining files for processing")
        input_dataset = DocumentDataset.read_parquet(
            files, backend="cudf", add_filename=True
        )
        result_dataset = classifier(dataset=input_dataset)
        result_dataset.to_parquet(qm_base_path, write_to_filename=True)
    else:
        logging.info("Nothing to be done. All files are already processed.")

    logging.info(f"Done in {time.time() - t0}")



# Copyright (c) 2025, NVIDIA CORPORATION.  All rights reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import warnings
from typing import List, Optional, Union

import dask.dataframe as dd


def deduplicate_groups(
    duplicates: dd.DataFrame, group_field: Optional[str], perform_shuffle: bool
) -> dd.DataFrame:
    if group_field is None:
        return duplicates

    if perform_shuffle:
        # Redistribute data across partitions so that all duplicates are in same partition
        duplicates_shuffled = duplicates.shuffle(on=[group_field], ignore_index=True)
    else:
        duplicates_shuffled = duplicates

    duplicates_to_remove = (
        duplicates_shuffled
        # For each partition, keep only the duplicated rows (excluding first occurrence)
        .map_partitions(lambda x: x[x[group_field].duplicated(keep="first")]).drop(
            columns=group_field
        )
    )
    return duplicates_to_remove


def left_anti_join(
    left: dd.DataFrame,
    right: dd.DataFrame,
    left_on: Union[str, List[str]],
    right_on: Union[str, List[str]],
) -> dd.DataFrame:
    assert left_on != right_on, "left_on and right_on cannot be the same"

    merge = left.merge(
        right=right,
        how="left",
        broadcast=True,  # Broadcast smaller DataFrame to all partitions
        left_on=left_on,
        right_on=right_on,
    )

    # This effectively removes all rows that were not in duplicates_to_remove
    removed_result = merge[merge[right_on].isna()].drop(columns=[right_on])
    return removed_result


def remove_duplicates(
    left: dd.DataFrame,
    duplicates: dd.DataFrame,
    id_field: str,
    group_field: Optional[str] = None,
    perform_shuffle: bool = False,
) -> dd.DataFrame:
    left_npartitions = left.optimize().npartitions
    right_npartitions = duplicates.optimize().npartitions
    if left_npartitions < right_npartitions:
        msg = (
            f"The number of partitions in `dataset` ({left_npartitions}) is less than "
            f"the number of partitions in the duplicates ({right_npartitions}). "
            "This may lead to a shuffle join. Repartitioning right dataset to match left partitions."
            "To control this behavior, call identify_duplicates and removal as two separate steps"
        )
        warnings.warn(msg)
        duplicates = duplicates.repartition(npartitions=left_npartitions)

    # Create a new column name for temporary ID storage during merge
    new_id_field = f"{id_field}_new"

    duplicates_to_remove = (
        deduplicate_groups(duplicates, group_field, perform_shuffle)
        # Rename the ID field to avoid conflicts in the upcoming merge
        .rename(columns={id_field: new_id_field})[[new_id_field]]
    )

    return left_anti_join(
        left=left,
        right=duplicates_to_remove,
        left_on=id_field,
        right_on=new_id_field,
    )
